% --- Source file: man/plot_xy_hist.Rd ---
\name{plot_xy_hist}
\alias{plot_xy_hist}
\title{
Scatter plot of histograms}
\description{
Performs a scatter plot where each "point" is a histogram
}
\usage{
plot_xy_hist(veclist, x, y, nbars=5, force=0, xscale=1, yscale=1, 
             xlab="", ylab="", title="", col="black")
}
\arguments{
  \item{veclist}{List of vectors from which each histogram is determined. The vectors can be of different lengths.}
  \item{x}{Vector of x-coordinates for each histogram. The length of this vector must be equal to length(\code{veclist}).}
  \item{y}{Vector of y-coordinates for each histogram. The length of this vector must be equal to length(\code{veclist}).}
  \item{nbars}{Number of bars in each histogram (the \code{breaks} option in \code{\link[graphics]{hist}}) }
  \item{force}{0 or 1 to force the number of bars in each histogram to be \code{nbars}.}
  \item{xscale}{Scaling factor for the width of each histogram.}
  \item{yscale}{Scaling factor for the height of each histogram.}
  \item{xlab}{X-axis label}
  \item{ylab}{Y-axis label}
  \item{title}{Title of plot}
  \item{col}{Color of each histogram}
}
\value{
  The returned value is a list of sublists of information about each histogram. The order is the same as \code{veclist}.
}
\details{
  The lower left point of the histogram defined by \code{veclist[[i]]} has coordinates \code{(x[i], y[i])}.
  The options \code{nbars}, \code{force}, and \code{col} are allowed to be vectors of the same length as \code{x}.
}

\author{Kai Yu}
\examples{

  set.seed(123)
  nr <- 10000
  nc <- 50
  dlist <- list()
  for (i in 1:nc) dlist[[i]] <- rnorm(ceiling(1000+nr*runif(1)))
  x <- 2*runif(nc) + 10*runif(nc)
  y <- 2*runif(nc) + 5*runif(nc)

  ret <- plot_xy_hist(dlist, x, y, xscale=0.25, yscale=0.5)

}
\seealso{\code{\link{heatMap}}}