#include "./header.h"


void print_iVec(vec, n, name)
int *vec;
int n;
char name[10];
{
  int i;
  printf("%s \n", name);
  for (i=0; i<n; i++) {
    printf(" %d ", vec[i]);
  }
  printf("\n \n");
}

void print_dVec(vec, n, name)
double *vec;
int n;
char name[10];
{
  int i;
  printf("%s \n", name);
  for (i=0; i<n; i++) {
    printf(" %g ", vec[i]);
  }
  printf("\n \n");
}


/* Sorting function, x is changed, sort from samll to large*/
void quicksort(int start, int stop, double *x, int *cvec)
{
 int i, j, k;
 double temp, median;
 int tempd;


  while (start < stop) {
    /*
    ** first-- if the list is short, do an ordinary insertion sort
    */
    if ((stop-start)<11) {
	for (i=start+1; i<=stop; i++) {
	    temp = x[i];
	    tempd= cvec[i];
	    j=i-1;

	    while (j>=start && (x[j]>temp)) {
		x[j+1] = x[j];
		cvec[j+1] = cvec[j];
		j--;
		}
	    x[j+1] = temp;
	    cvec[j+1]  = tempd;
	    }
	return;
	}

    /*
    ** list is longer -- split it into two
    **  I use the median of 3 values as the split point
    */
    i=start;
    j=stop;
    k = (start + stop)/2;

    median = x[k];
    if (x[i] >= x[k]) {      /* one of j or k is smallest */
	if (x[j] > x[k]) {   /* k is smallest */
	    if (x[i] > x[j])  median = x[j];
	    else median= x[i];
	    }
	}
    else {
	if (x[j] < x[k]) {
	    if (x[i] > x[j]) median = x[i];
	    else median = x[j];
	    }
	}

    /* 
    **  Now actually do the partitioning 
    **   Because we must have at least one element >= median, "i"
    **   will never run over the end of the array.  Similar logic
    **   applies to j.
    ** A note on the use of "<" rather than "<=".  If a list has lots
    **   of identical elements, e.g. 80/100 are "3.5", then we will
    **   often go to the swap step with x[i]=x[j]=median.  But we will
    **   get the pointers i and j to meet approximately in the middle of
    **   the list, and that is THE important condition for speed in a
    **   quicksort.
    **   
    */
    while (i<j) {
	/*
	** top pointer down till it points at something too large
	*/
	while (x[i] < median) i++;

	/*
	** bottom pointer up until it points at something too small
	*/
	while(x[j] > median) j--;

	if (i<j) {
	    if (x[i] > x[j]) {  /* swap */
		temp = x[i];
		x[i] = x[j];
		x[j] = temp;
		tempd= cvec[i];   cvec[i] =cvec[j];  cvec[j] =tempd;
		}
	    i++; j--;
	    }
	}

    /*
    ** The while() step helps if there are lots of ties.  It will break
    **  the list into 3 parts: < median, ==median, >=median, of which only
    **  the top and bottom ones need further attention.
    ** The ">=" is needed because i may be  == to j
    */
    while (x[i] >= median && i>start) i--;
    while (x[j] <= median && j<stop ) j++;

    /*
    ** list has been split, now do a recursive call
    **   always recur on the shorter list, as this keeps the total
    **       depth of nested calls to less than log_base2(n).
    */
    if ((i-start) < (stop-j)) { /* top list is shorter */
	if ((i-start)>0) quicksort(start,i, x, cvec);
	start =j; 
	}

    else {    /* bottom list is shorter */
	if ((stop -j)>0) quicksort(j,stop, x, cvec);
	stop=i; 
	}
     }
}

/* Function to get the ranks of a double vector */
void get_ranks(vec, n, rank)
double *vec;
int n;
int *rank; /* Output ranks */
{
  int i, *pi1, tempi[n];
  double temp[n], *p1, *p2;

  /* Copy vec to temp */
  for (i=0, p1=vec, p2=temp, pi1=tempi; i<n; i++, p1++, p2++, pi1++) {
    *p2 = *p1;
    *pi1 = i;
  }

  quicksort(0, n-1, temp, tempi);

  /* Get the correct order for the vector */
  for (i=0, pi1=tempi; i<n; i++, pi1++) rank[*pi1] = i;

} 

/*************************************************************/
/* Functions for allocating matrices */
/*************************************************************/
int ** intMatrix_alloc(nrow, ncol)
int nrow, ncol;
{
  int **mat, i;

  mat = (int **) malloc(nrow*sizeof(int *));
  CHECK_MEM(mat);
  for (i=0; i<nrow; i++) {
    mat[i] = (int *) malloc(ncol*sizeof(int));
    CHECK_MEM(mat);
  }
  return(mat);
}

void intMatrix_free(mat, nrow)
int **mat, nrow;
{
  int i;
  
  for (i=0; i<nrow; i++) free(mat[i]);
  free(mat);
}

double ** dblMatrix_alloc(nrow, ncol)
int nrow, ncol;
{
  double **mat;
  int i;

  mat = (double **) malloc(nrow*sizeof(double *));
  CHECK_MEM(mat);
  for (i=0; i<nrow; i++) {
    mat[i] = (double *) malloc(ncol*sizeof(double));
    CHECK_MEM(mat);
  }
  return(mat);
}

void dblMatrix_free(mat, nrow)
double **mat;
int nrow;
{
  int i;
  
  for (i=0; i<nrow; i++) free(mat[i]);
  free(mat);
}

/*******************************************************/
/* Prior functions for the number of clusters */
/* not used for the current version */
/*******************************************************/
int pri_gen_num_of_cluster(K_MAX)
int K_MAX;
{
  int k;
  k = uniform_int(1, K_MAX);
  return(k);
}
double pri_pd_num_of_cluster(k, K_MAX)
     int k, K_MAX;
{
  return(1.0 / K_MAX );
}

/*******************************************************/
/* Prior functions for psi interaction based on uniform prior*/
/*******************************************************/
double pri_gen_psi_unif(PSI_MIN, PSI_MAX)
double PSI_MIN, PSI_MAX;
{
  double u, ret;

  u = runif(0.0, 1.0);
  ret = PSI_MIN + (PSI_MAX - PSI_MIN)*u;
 
  return(ret);
}
double pri_pd_psi_unif(psi, PSI_MIN, PSI_MAX)
double psi, PSI_MIN, PSI_MAX;
{
  return(1.0 / (PSI_MAX - PSI_MIN));
}

/*******************************************************/
/* Prior functions for psi interaction based on gamma prior*/
/*******************************************************/
double pri_gen_psi_gamma(PSI_k, PSI_theta)
double PSI_k, PSI_theta;
{
  double ret;

  /*ret = gsl_ran_gamma(rngptr, PSI_k, PSI_theta);*/
  ret = rgamma(PSI_k, PSI_theta);
 
  return(ret);
}
double pri_pd_psi_gamma(psi, PSI_k, PSI_theta)
double psi, PSI_k, PSI_theta;
{
  /*return(gsl_ran_gamma_pdf(psi, PSI_k, PSI_theta));*/
  return(dgamma(psi, PSI_k, PSI_theta, 0));
}

/*******************************************************/
/* Prior functions for the alpha */
/*******************************************************/
static void pri_gen_alpha_unif(alpha, k, alpha_MIN, alpha_MAX)
double *alpha;
int k;
double alpha_MIN, alpha_MAX;
{
  double u, *ptrd = alpha;
  int i;

  for (i=0; i<k; i++) {
    u = runif(0.0, 1.0);
    *ptrd++ = alpha_MIN + (alpha_MAX - alpha_MIN)*u;
  }

}

static double pri_pd_alpha_unif(alpha, alpha_MIN, alpha_MAX)
double alpha, alpha_MIN, alpha_MAX;
{
  return(1.0 / (alpha_MAX - alpha_MIN));
}

static void pri_gen_alpha_norm(alpha, k, mu, sigma2)
double *alpha;
int k;
double mu, sigma2;
{
  rnormal_vec(mu, sigma2, alpha, k);
}

static double pri_pd_alpha_norm(alpha, mu, sigma2)
double alpha, mu, sigma2;
{
  double u, ret;
  u = alpha - mu;

  /*return(gsl_ran_gaussian_pdf(u, sqrt(sigma2)));*/
  ret = dnorm(u, 0.0, sqrt(sigma2), 0);
  return(ret);
}

void pri_gen_alpha(alpha, k, alpha_MIN, alpha_MAX, method, mu, sigma2)
double *alpha;
int k;
double alpha_MIN, alpha_MAX;
int method;
double mu, sigma2;
{

  if (method == 0) {
    pri_gen_alpha_unif(alpha, k, alpha_MIN, alpha_MAX);
  } else {
    pri_gen_alpha_norm(alpha, k, mu, sigma2);
  }

}

double pri_pd_alpha(alpha, alpha_MIN, alpha_MAX, method, mu, sigma2)
double alpha, alpha_MIN, alpha_MAX, mu, sigma2;
int method;
{
  double ret;

  if (method == 0) {
    ret = pri_pd_alpha_unif(alpha, alpha_MIN, alpha_MAX);
  } else {
    ret = pri_pd_alpha_norm(alpha, mu, sigma2);
  }
  return(ret);
}

/*******************************************************/
/* Prior functions for beta (log-odds ratio) */
/*******************************************************/
static void pri_gen_beta_unif(beta, k, BETA_MIN, BETA_MAX)
double *beta;
int k;
double BETA_MIN, BETA_MAX;
{
  double u, *ptrd = beta;
  int i;

  for (i=0; i<k; i++) {
    u = runif(0.0, 1.0);
    *ptrd++ = BETA_MIN + (BETA_MAX - BETA_MIN)*u;
  }

}
static double pri_pd_beta_unif(beta, BETA_MIN, BETA_MAX)
double beta, BETA_MIN, BETA_MAX;
{
  return(1.0 / (BETA_MAX - BETA_MIN));
}

static void pri_gen_beta_norm(beta, k, mu, sigma2)
double *beta;
int k;
double mu, sigma2;
{
  rnormal_vec(mu, sigma2, beta, k);
}

static double pri_pd_beta_norm(beta, mu, sigma2)
double beta, mu, sigma2;
{
  double u, ret;
  u = beta - mu;

  /*return(gsl_ran_gaussian_pdf(u, sqrt(sigma2)));*/
  ret = dnorm(u, 0.0, sqrt(sigma2), 0);
  return(ret);
}

void pri_gen_beta(beta, k, beta_MIN, beta_MAX, method, mu, sigma2)
double *beta;
int k;
double beta_MIN, beta_MAX;
int method;
double mu, sigma2;
{

  if (method == 0) {
    pri_gen_beta_unif(beta, k, beta_MIN, beta_MAX);
  } else {
    pri_gen_beta_norm(beta, k, mu, sigma2);
  }

}

double pri_pd_beta(beta, beta_MIN, beta_MAX, method, mu, sigma2)
double beta, beta_MIN, beta_MAX, mu, sigma2;
int method;
{
  double ret;

  if (method == 0) {
    ret = pri_pd_beta_unif(beta, beta_MIN, beta_MAX);
  } else {
    ret = pri_pd_beta_norm(beta, mu, sigma2);
  }
  return(ret);
}

/*******************************************************/
/* Prior functions for tao, the coef for covariates */
/*******************************************************/
void pri_gen_tao_unif(tao, k, TAO_MIN, TAO_MAX)
double *tao;
int k;
double *TAO_MIN, *TAO_MAX;
{
  double u, *ptrd = tao;
  int i;

  for (i=0; i<k; i++) {
    u = runif(0.0, 1.0);
    *ptrd++ = TAO_MIN[i] + (TAO_MAX[i] - TAO_MIN[i])*u;
  }

}

double pri_pd_tao_unif(tao, TAO_MIN, TAO_MAX)
double tao, TAO_MIN, TAO_MAX;
{
  return(1.0 / (TAO_MAX - TAO_MIN));
}

static void pri_gen_tao_norm(tao, k, mu, sigma2)
double *tao;
int k;
double *mu, *sigma2;
{
  int i;

  for (i=0; i<k; i++) rnormal_vec(mu[i], sigma2[i], &tao[i], 1);
}

static double pri_pd_tao_norm(tao, mu, sigma2)
double tao, mu, sigma2;
{
  double u, ret;
  u = tao - mu;

  /*return(gsl_ran_gaussian_pdf(u, sqrt(sigma2)));*/
  ret = dnorm(u, 0.0, sqrt(sigma2), 0);
  return(ret);
}

void pri_gen_tao(tao, k, tao_MIN, tao_MAX, method, mu, sigma2)
double *tao;
int k;
double *tao_MIN, *tao_MAX;
int method;
double *mu, *sigma2;
{

  if (method == 0) {
    pri_gen_tao_unif(tao, k, tao_MIN, tao_MAX);
  } else {
    pri_gen_tao_norm(tao, k, mu, sigma2);
  }

}

double pri_pd_tao(tao, tao_MIN, tao_MAX, method, mu, sigma2)
double tao, tao_MIN, tao_MAX, mu, sigma2;
int method;
{
  double ret;

  if (method == 0) {
    ret = pri_pd_tao_unif(tao, tao_MIN, tao_MAX);
  } else {
    ret = pri_pd_tao_norm(tao, mu, sigma2);
  }
  return(ret);
}


/*******************************************************/
/* Function to assign alloc_z from 0 to k-1 */
/* this function return a z with exactly k cluster */
/*******************************************************/

void ini_gen_alloc(z, n, k)
int *z;
int n; /* Number of subjects */
int k; /* Number of clusters */
{
  int i, *ptri=z, clusFlag[k], clus, sum, km1=k-1, iter, maxiter=100, ok_flag=0;

  if (n < k) {
    printf("\n ERROR: Number of subjects < number of clusters \n");
    exit(1);
  }

  for (iter=0; iter<maxiter; iter++) {
    /* Initialize to 0 */
    ivecinit(clusFlag, k, 0);

    for (i=0, ptri=z; i<n; i++, ptri++) {
      /*clus = (int) gsl_rng_uniform_int(rngptr, k);*/
      clus = uniform_int(0, km1);
      *ptri = clus;
      clusFlag[clus] = 1; 
    }
    sum = 0;
    for (i=0, ptri=clusFlag; i<k; i++, ptri++) sum += *ptri;
    if (sum == k) {
      ok_flag = 1;
      break;
    }
  }

  /* If all clusters were not assigned print a warning. Set the first k elements of z to be 0, ..., k-1 */
  if (!ok_flag) {
    printf("WARNING: Not all clusters were randomly assigned in ini_gen_alloc\n");
    for (i=0, ptri=z; i<k; i++, ptri++) *ptri = i;
  }

}


/*******************************************************/
/* Function for computing U(z) */
/* this function does not need to be changed, we just treat the num_group as num_subj */
/*******************************************************/
double calc_Uz(sym_mat, z_vector, num_subj, neighbor_id, num_neighbor)
double **sym_mat; /* Symmetric matrix. Assume matrix is allocated by rows (num_subj x num_subj) */
int *z_vector;  /* num_subj x 1 */
int num_subj;
int **neighbor_id, *num_neighbor; 
{
  int i, j, z_i, n, neigh, *neigh_ids;
  double sum=0., *ptrd;
  
  for (i=0; i<num_subj; i++) {
    n = num_neighbor[i];
    if (n) {
      z_i = z_vector[i];
      ptrd = sym_mat[i];
      neigh_ids = neighbor_id[i];
      for (j=0; j<n; j++) {
        neigh = neigh_ids[j];
        if ((neigh > i) && (z_i == z_vector[neigh])) sum += ptrd[neigh];
      }
    }
  }

  return(sum);
}

/******************************************************/
/* Functions for reading the data */
/******************************************************/

static int count_strinstr(big, little)
const char *big;
const char *little;
{
const char *p;
int count = 0;
size_t lil_len = strlen(little);

  /* you decide what to do here */
  if (lil_len == 0) return 0;

  p = strstr(big, little);
  while (p) {
    count++;
    p = strstr(p + lil_len, little);
  }
  return count;
}

/* Function for opening a data set in read mode */
FILE * open_r(datafile)
const char *datafile;
{
  FILE *fpt;

  fpt = fopen(datafile, "r");
  if (fpt == 0) {
    printf("\n Error opening file: %s \n", datafile);
    exit(0);
  }

  return(fpt);
}



/* Function for getting the number of columns in a (space delimited) data file */
int getNcols(datafile)
const char *datafile;
{
  FILE *fpt;
  int i = 0;
  char line[45000];

  fpt = open_r(datafile);
  fgets(line, sizeof(line), fpt);
  fclose(fpt);
  i = count_strinstr(line, " ") + 1;

  return(i);
}


/* Function for getting the number of rows in a data file */
int getNrows(datafile)
const char *datafile;
{
  FILE *fpt;
  int i = 0;
  char line[80];

  fpt = open_r(datafile);
  while(fgets(line, sizeof(line), fpt) != NULL) i++;
  fclose(fpt);
  return(i);
}

/* Function for reading data */
static void read_data(datafile, datastr)
const char *datafile;
struct data *datastr;
{
  FILE *fpt;
  int i, j, temp, *ptr_ds, z1, z3, *ptr_grp, ncovars=datastr->num_of_covars;
  float x, z2;
  double *ptr_x;
  double *ptrd, **covars=datastr->covars;
  
  fpt = open_r(datafile);
  i = 1;
  ptr_ds = datastr->disease_status;
  ptr_x  = datastr->x;
  ptr_grp = datastr->grp_id;

  while(i<=datastr->num_of_subj && !feof(fpt)) 
  {   
    /* First column is disease status, second genotype, third group id */
    temp = fscanf(fpt, "%d %g %d", &z1, &z2, &z3);
    if (temp == EOF) break;
   
    /* Read in the covariates, row is subject, col is variable */
    if (ncovars) {
      for (j=0, ptrd=covars[i-1]; j<ncovars; j++, ptrd++) {
        temp = fscanf(fpt, "%g", &x);
        *ptrd = (double) x; 
      }
    }

    *ptr_ds++ = z1;
    *ptr_x++ =  (double)z2;
    *ptr_grp++ =  z3;
    i++;
  }
  fclose(fpt);
}

/* Function for reading the similarity matrix */
static void read_matrix(datafile, datastr)
const char *datafile;
struct data *datastr;
{
  FILE *fpt;
  float z;
  int i, j, temp, n;
  double **ptr_mat;
  
  fpt = open_r(datafile);
  i = 1;
  ptr_mat = datastr->simi_mat;
  n = datastr->num_of_grp;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      temp = fscanf(fpt, "%f", &z);
      if (temp == EOF) {
        printf(" \n Error with similarity matrix file \n");
        exit(0);
      }
      ptr_mat[i][j] = (double) z;
    }
  }
  fclose(fpt);
}

/************************************************************/
/* Functions for checking the input data */
/***********************************************************/

static int checkMatrix(mat, n)
double **mat;
int n;
{
  int i, j, ret = 0;
  

  for (i=0; i<n; i++) {
    if (fabs(mat[i][i]) > DOUBLE_TOL) return(1); 
    for (j=0; j<n; j++) {
      if (fabs(mat[i][j] - mat[j][i]) > DOUBLE_TOL) return(1); 
    }
  }
  return(ret);
}

/* Return the number of groups */
static int checkData(datastr)
struct data *datastr;
{
  int i, d, *ptri=datastr->disease_status, n=datastr->num_of_subj;
  int *ptr_grp=datastr->grp_id, ngrp=0, *groups;

  for (i=0; i<n; i++) {
    d = *ptri++;
    if ((d != 0) && (d != 1)) {
      printf("ERROR: disease status must be 0/1 \n");
      exit(0);
    }
    d = *ptr_grp++;
    ngrp = MAX(ngrp, d);
  }
  ngrp++;

  /* Check the groups from 0 to ngrp - 1 */
  groups = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(groups);
  ivecinit(groups, ngrp, 0);
  for (i=0, ptr_grp=datastr->grp_id; i<n; i++, ptr_grp++) groups[*ptr_grp] = 1;
  d = 0;
  for (i=0, ptri=groups; i<ngrp; i++, ptri++) d += *ptri;
  if (d != ngrp) {
    printf("\n ERROR: groups must be coded 0, 1, ..., H-1 \n");
    exit(1);
  }

  return(ngrp);
}

/* Function to assign neighbors using the similarity matrix */
void assignNeighbors(datastr)
struct data *datastr;
{
  int i, j, ngrp=datastr->num_of_grp, *pint, temp[datastr->num_of_grp], n, *num, **ids;
  double **sim=datastr->simi_mat, *pd;

  /* Allocate memory */
  datastr->num_neighbor = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->num_neighbor);
  datastr->neighbor_id = (int **) malloc(ngrp*sizeof(int *));
  CHECK_MEM(datastr->neighbor_id);

  num = datastr->num_neighbor;
  ids = datastr->neighbor_id;
  for (i=0; i<ngrp; i++) {
    pd = sim[i];
    n = 0;
    /* First determine how many neighbors there are */
    for (j=0; j<ngrp; j++) {
      if ((i != j) && (pd[j] > ZERO_THRS)) {
        temp[n] = j;
        n++; 
      }
    }
    /* Allocate memory and asign */
    num[i] = n;
    if (n) {
      ids[i] = (int *) malloc(n*sizeof(int));
      CHECK_MEM(ids[i]);
      pint = ids[i];
      for (j=0; j<n; j++) pint[j] = temp[j]; 
    }
  }

} /* END: assignNeighbors */

/* to initial the data, the some parameters including the number of cluster need to be specified before */
void init(datastr, datafile, sim_mat_file)
struct data *datastr;
const char *datafile;
const char *sim_mat_file;
{
  int i, n, k, ngrp, j, *ptri, *ptri2, ncov;
  /*const gsl_rng_type *T;*/
  int Kp1, K_MAX;
  double *b, *c, *d;
  int n_type1[datastr->num_of_cluster];

  /* Get the number of clusters */
  k = datastr->num_of_cluster;
  K_MAX = datastr->k_max;
  Kp1 = K_MAX + 1;

  /* For random numbers */
  GetRNGstate();

  /* Get the number of subjects */
  datastr->num_of_subj = getNrows(datafile);
  n = datastr->num_of_subj;

  /* Get the number of covariates */
  datastr->num_of_covars = getNcols(datafile) - 3;

  /* Allocate memory */
  datastr->x = (double *) malloc(n*sizeof(double));
  CHECK_MEM(datastr->x);
  datastr->disease_status = (int *) malloc(n*sizeof(int));
  CHECK_MEM(datastr->disease_status);

  datastr->offset = (double *) malloc(n*sizeof(double));
  CHECK_MEM(datastr->offset);
  vecinit(datastr->offset, n, 0.0);  

  datastr->b = (double *) malloc(K_MAX*sizeof(double));
  CHECK_MEM(datastr->b);
  datastr->c = (double *) malloc(K_MAX*sizeof(double));
  CHECK_MEM(datastr->c);
  datastr->d = (double *) malloc(K_MAX*sizeof(double));
  CHECK_MEM(datastr->d);

  datastr->alpha = (double *) malloc(Kp1*sizeof(double));
  CHECK_MEM(datastr->alpha);
  datastr->beta = (double *) malloc(Kp1*sizeof(double));
  CHECK_MEM(datastr->beta);
  datastr->grp_id = (int *) malloc(n*sizeof(int));
  CHECK_MEM(datastr->grp_id);

  if (datastr->num_of_covars) {
    /* Check the tao min, max, mu, sigma2 vectors */
    ncov = datastr->num_of_covars;

    if (datastr->method_tao == 0) {
      /* Uniform */
      if ((ncov != datastr->tao_min_n0) || (ncov != datastr->tao_max_n0)) {
        printf("ERROR with tao_min and/or tao_max options\n");
        exit(1);
      }
      if (datastr->tao_mu) {
        free(datastr->tao_mu);
        datastr->tao_mu = NULL;
      }
      if (datastr->tao_sigma2) {
        free(datastr->tao_sigma2);
        datastr->tao_sigma2 = NULL;
      }
    } else {
      /* Normal */
      if ((ncov != datastr->tao_mu_n0) || (ncov != datastr->tao_sigma2_n0)) {
        printf("ERROR with tao_mu and/or tao_sigma2 options\n");
        exit(1);
      }
      if (datastr->tao_min) {
        free(datastr->tao_min);
        datastr->tao_min = NULL;
      }
      if (datastr->tao_max) {
        free(datastr->tao_max);
        datastr->tao_max = NULL;
      }
    }
    datastr->covars = dblMatrix_alloc(n, datastr->num_of_covars);
    datastr->tao = (double *) malloc(datastr->num_of_covars*sizeof(double));
    CHECK_MEM(datastr->tao);
  } else {
    if (datastr->tao_mu) {
      free(datastr->tao_mu);
      datastr->tao_mu = NULL;
    }
    if (datastr->tao_sigma2) {
      free(datastr->tao_sigma2);
      datastr->tao_sigma2 = NULL;
    }
    if (datastr->tao_min) {
      free(datastr->tao_min);
      datastr->tao_min = NULL;
    }
    if (datastr->tao_max) {
      free(datastr->tao_max);
      datastr->tao_max = NULL;
    }
  }

  /* assign b, c, and d, with b for split, d for merge, c for z update  */
  b = datastr->b;
  c = datastr->c;
  d = datastr->d;

  b[0] = 1.0;
  c[0] = 0.0;
  d[0] = 0.0;
  b[K_MAX-1] = 0.0;
  c[K_MAX-1] = 0.5;
  d[K_MAX-1] = 0.5;
  for (i=1; i<K_MAX-1; i++) {
    b[i] = 1.0/3.0;
    c[i] = 1.0/3.0;
    d[i] = 1- b[i] - c[i];
  }


  /* Read in x, disease status, and group id and covariates*/
  read_data(datafile, datastr);

  /* Check disease status and return the number of groups */
  ngrp = checkData(datastr);
  datastr->num_of_grp = ngrp;

  /* number of subj in each grp */
  datastr->n_subj_grp = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->n_subj_grp);

  /* number of cases in each group */
  datastr->n_case_grp = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->n_case_grp);

  datastr->n_control_grp = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->n_control_grp);
 
  /* number of grps in each cluster */
  datastr->n_grp_cluster = (int *) malloc(datastr->num_of_cluster*sizeof(int));
  CHECK_MEM(datastr->n_grp_cluster);

  /* number of cases in each cluster */
  datastr->n_case_cluster = (int *) malloc(datastr->num_of_cluster*sizeof(int));
  CHECK_MEM(datastr->n_case_cluster);

  datastr->n_control_cluster = (int *) malloc(datastr->num_of_cluster*sizeof(int));
  CHECK_MEM(datastr->n_control_cluster);


  /* Get the group counts */
  cluster_counts(datastr->n_subj_grp, ngrp, datastr->grp_id, n);

  /* Get the case counts */ 
   cluster_counts_case(datastr->n_case_grp, ngrp, datastr->grp_id, n,
		       datastr->disease_status, 1);
  
   cluster_counts_case(datastr->n_control_grp, ngrp, datastr->grp_id, n,
		       datastr->disease_status, 0);

  /* Allocate memory for groups */
  datastr->subj_grp = (int **) malloc(ngrp*sizeof(int *));
  CHECK_MEM(datastr->subj_grp);
  for (i=0; i<ngrp; i++) {
    datastr->subj_grp[i] = (int *) malloc(datastr->n_subj_grp[i]*sizeof(int));
    CHECK_MEM(datastr->subj_grp[i]);

    /* Assign the list of subjects in each group */
    ptri2 = datastr->subj_grp[i];
    for (j=0, ptri=datastr->grp_id; j<n; j++, ptri++) {
      if (*ptri == i) *ptri2++ = j; /* check with Bill */
    }
  }

  /* Allocate memory for the z vector, w0, w matrix, and similarity matrix */
  datastr->alloc_z = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->alloc_z);
  datastr->w0 = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(datastr->w0);
  datastr->auxillary_sample = intMatrix_alloc(datastr->w_m, ngrp);
  datastr->simi_mat = dblMatrix_alloc(ngrp, ngrp);

  /* Read in the similarity matrix */
  read_matrix(sim_mat_file, datastr);

  /* Check that the matrix is symmetric with zero on diagonal */
  if (checkMatrix(datastr->simi_mat, ngrp)) {
    printf("\n ERROR: Similarity matrix should be symmetric with 0 on the diagonal \n");
    exit(0);
  }

  /* PSI interaction */
  /* for the inital, we can start from uniform, but can still set its prior as other distribution */
  if (datastr->method_psi == 2) {
    j = sample1(datastr->prob_psi_grid, datastr->n_psi_grid);
    datastr->psi_position = j;
    datastr->psi_interaction = datastr->psi_grid[j];
  } else {
    datastr->psi_interaction = pri_gen_psi_unif(datastr->psi_min, datastr->psi_max);
  }

  /* Alpha */
  pri_gen_alpha(datastr->alpha, datastr->num_of_cluster, datastr->alpha_min, datastr->alpha_max,
                datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2);

  /* Beta */
  pri_gen_beta(datastr->beta, datastr->num_of_cluster, datastr->beta_min, datastr->beta_max,
               datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2);

  /* Tao */
  if (datastr->num_of_covars) {
    pri_gen_tao(datastr->tao, datastr->num_of_covars, datastr->tao_min, datastr->tao_max,
                datastr->method_tao, datastr->tao_mu, datastr->tao_sigma2);
    for (i=0; i < n; i++) {
	for (j = 0; j < datastr->num_of_covars; j++) {
          (datastr->offset)[i] = (datastr->offset)[i] + (datastr->tao)[j] * (datastr->covars)[i][j];
       }
    }
  }

  /* alloc_z */
  j = 0;
  for (i=0; i < 100; i++)
    {
      ini_gen_alloc(datastr->alloc_z, ngrp, datastr->num_of_cluster);
      /* check whether z is acceptable: each cluster has the required number*/
      /* of grps */
      cluster_counts(datastr->n_grp_cluster, datastr->num_of_cluster,
		     datastr->alloc_z, ngrp);
    
      j = accept_or_not_grp(datastr->n_grp_cluster, datastr->num_of_cluster,
			datastr->cluster_size); 
      if ( j ==1)
	{
	  break;
	}
    }
   
  if (j == 0)
    {
      printf("\n ERROR: can't find a eligible starting allocation point \n");
      exit(1);      
    }

  /* rerun for w0 */
  j = 0;
  for (i=0; i < 100; i++)
    {
      ini_gen_alloc(datastr->w0, ngrp, datastr->num_of_cluster);
      /* check whether z is acceptable: each cluster has the required number*/
      /* of grps */
      cluster_counts(n_type1, datastr->num_of_cluster,
		     datastr->w0, ngrp);
    
      j = accept_or_not_grp(n_type1, datastr->num_of_cluster,
			datastr->cluster_size); 
      if ( j ==1)
	{
	  break;
	}
    }
   
  if (j == 0)
    {
      printf("\n ERROR: can't find a eligible starting allocation point \n");
      exit(1);      
    }


 
  /* Allocate memory and assign neighbors */
  assignNeighbors(datastr);

  /* Initialize the auxillary sample w1, ..., wm using k_max, the first inital point is w0*/
  update_all_w(datastr, datastr->w0, datastr->auxillary_sample);

  /* Check the allocation and recode if necessary */
  /*check_z(datastr);*/

}


/*********************************************************/
/* Function to free memory */
/**********************************************************/
void free_memory(datastr)
struct data *datastr;
{
  int i, *pint, **p2;

  /*gsl_rng_free(datastr->rngptr);*/
  free(datastr->x);
  free(datastr->disease_status);
  free(datastr->alloc_z);
  free(datastr->w0);
  free(datastr->alpha);
  free(datastr->beta);
  free(datastr->grp_id);
  free(datastr->offset);
  free(datastr->b);
  free(datastr->c);
  free(datastr->d);

  dblMatrix_free(datastr->simi_mat, datastr->num_of_grp);
  intMatrix_free(datastr->auxillary_sample, datastr->w_m);
  free(datastr->n_subj_grp);
  free(datastr->n_case_grp); 
  free(datastr->n_control_grp);
  free(datastr->n_grp_cluster);
  free(datastr->n_case_cluster); 
  free(datastr->n_control_cluster);

  intMatrix_free(datastr->subj_grp, datastr->num_of_grp);
  if (datastr->num_of_covars) 
    {
      dblMatrix_free(datastr->covars, datastr->num_of_subj);
      free(datastr->tao);
      if (datastr->tao_min) free(datastr->tao_min);
      if (datastr->tao_max) free(datastr->tao_max);
      if (datastr->tao_mu) free(datastr->tao_mu);
      if (datastr->tao_sigma2) free(datastr->tao_sigma2);
    }

  pint = datastr->num_neighbor;
  p2   = datastr->neighbor_id;
  for (i=0; i<datastr->num_of_grp; i++) {
    if (pint[i]) free(p2[i]);
  }
  free(p2);
  free(pint);

  if (datastr->psi_grid) free(datastr->psi_grid);
  if (datastr->prob_psi_grid) free(datastr->prob_psi_grid);
  if (datastr->normalizing_psi_grid) free(datastr->normalizing_psi_grid);

  PutRNGstate();
}


/***********************************************************/
/* Random number generation */
/***********************************************************/

/* Generate a uniform integer between n1 and n2 inclusive */
int uniform_int(n1, n2) 
int n1, n2;
{
  int ret;
  
  ret = floor(runif(n1, n2+1));
  if (ret > n2) ret = n2;
  return(ret);
}

double rnormal(mu, sigma2)
double mu;
double sigma2;  /* Variance parmameter */
{
  double u;

  /* Pass in the standard deviation */
  /*u = mu + gsl_ran_gaussian(rngptr, sqrt(sigma2));*/
  u = rnorm(mu, sqrt(sigma2));

  return(u);
}

void rnormal_vec(mu, sigma2, outvec, n)
double mu;
double sigma2;  /* Variance parmameter */
double *outvec;
int n;
{
  double *ptrd = outvec, sd;
  int i;

  sd = sqrt(sigma2);
  /*for (i=0; i<n; i++) *ptrd++ = mu + gsl_ran_gaussian(rngptr, sd);*/
  for (i=0; i<n; i++) *ptrd++ = rnorm(mu, sd);
}

int bernoulli(prob)
double prob;
{
  if (prob > 1.0) prob = 1.0;

  /*return(gsl_ran_bernoulli(rngptr, prob));*/
  return(rbinom(1, prob));
}

/* Function to return an integer 0, ..., k-1 based on a vector of probs */
int sample1(probs, k)
double *probs; /* Must sum to 1 */
int k;
{
  double u, cumsum=0., *ptr=probs;
  int i;

  /* Let the probabilities be p1, p2, ..., pk.
     Generate a uniform random value and see which interval it lies in:
     [0, p1), [p1, p1+p2), ... , [p1+p2+...+pk-1, 1]
  */
  /*u = gsl_rng_uniform_pos(rngptr);*/
  u = runif(0.0, 1.0);
  for (i=0; i<k; i++) {
    cumsum += *ptr++;
    if (u < cumsum) break;
  }
  if (i == k) i = k - 1;
  return(i);
}

double uniform(a, b)
double a, b;
{
  double ret;

  /*
  u = gsl_rng_uniform_pos(rngptr);
  ret = a + (b - a)*u;
  */
  ret = runif(a, b);
  return(ret);
}


/**********************************************************/
/* Functions for vectors */
/**********************************************************/

/* Function to normalize a vector to unit length */
void vecnorm(invec, outvec, n)
double *invec;
double *outvec;
int n;
{
  double *ptrd = invec, sum=0.0, temp, *ptrd2 = outvec;
  int i;
  
  for (i=0; i<n; i++) {
    temp = *ptrd++;
    sum += temp*temp;
  }
  sum = 1./sqrt(sum);
  ptrd = invec;
  for (i=0; i<n; i++) *ptrd2++ = (*ptrd++)*sum;
  
}

/* Function to sum a double vector */
double vecsum(vec, n)
double *vec;
int n;
{
  double sum = 0.;
  
  while (n-- > 0) {
    sum += *vec++;
  } 
  return(sum);

}

/* Function to normalize a vector to sum to 1 */
void vecnormSum(invec, outvec, n)
double *invec;
double *outvec;
int n;
{
  double *ptrd = invec, sum, *ptrd2 = outvec;
  int i;
  
  sum = vecsum(invec, n);
  sum = 1./sum;
  for (i=0; i<n; i++) *ptrd2++ = (*ptrd++)*sum;
  
}

/* Function to check the bounds of a vector */
int checkVecBounds(vec, n, lower, upper)
double *vec;
int n;
double lower;
double upper;
{
  int i;
  double *ptrd = vec, temp;
  
  for (i=0; i<n; i++) {
    temp = *ptrd++;
    if ((temp <= lower) || (temp >= upper)) return(1); 
  }
  return(0);
}

/* Function to transform a double vector: outvec = vec1 + scale*vec2 */
void Vec1_p_cVec2(outvec, vec1, vec2, n, scale)
double *outvec, *vec1, *vec2, scale;
int n;
{
  int i;
  double *ptr1 = vec1, *ptr2 = vec2, *ptrout = outvec;

  for (i=0; i<n; i++) *ptrout++ = *ptr1++ + scale*(*ptr2++);
}

/* Function to initialize a double vector to a constant */
void vecinit(vec, n, value)
double *vec, value;
int n;
{
  double *ptr = vec;
  while (n-- > 0) {
    *ptr++ = value;
  }
}

/* Function to initialize an integer vector to a constant */
void ivecinit(vec, n, value)
int *vec, n, value;
{
  int *ptr = vec;
  while (n-- > 0) {
    *ptr++ = value;
  }
}

/* Function to initialize an integer vector to another vector */
void ivec_ivec_init(outvec, vec, n)
int *outvec, *vec;
int n;
{
  int *ptr1 = vec, *ptr2 = outvec;
  while (n-- > 0) {
    *ptr2++ = *ptr1++;
  }
}

/* Function to sum elements of an integer vecotr that are > m */
/*
static int ivecsum_gtm(vec, n, t_cut)
int *vec, n;
double t_cut;
{
  int *ptr=vec, temp, sum=0;

  while (n-- > 0) {
    temp = *ptr++;
    if (temp > t_cut) sum += temp;
  }
  return(sum);
}
*/

/* caluate the -2log(likelihood) given the current datastructure */
double cal_deviance(D, G, GRP, n, Z, alpha, beta)
int *D; /* Disease status */
double *G; /* Genotypes */
int *GRP; /* Group number */
int *Z; /* Cluster for each group */
double *alpha, *beta;
int n; /* length of D, G, and GRP */
{
  double  sum = 0.0, t_loc, beta_zi, alpha_zi;
  int i, zi, *ptrD, *ptrGRP, grpi ;
  double *ptrG, gi;
 

  for (i=0, ptrD=D, ptrG=G, ptrGRP=GRP; i<n; i++, ptrD++, ptrG++, ptrGRP++) {
      grpi = *ptrGRP;
      zi = Z[grpi]; 
      gi = *ptrG; 
      beta_zi = beta[zi]; /* beta remains fixed in the formula */
      alpha_zi = alpha[zi];
      
      t_loc = exp(alpha_zi + beta_zi * gi);
      if (*ptrD) {
         /* for a case */    
          sum += log(t_loc / (1.0 +t_loc));
      } else {
        /* for a control */
        sum += log(1.0 / (1.0 +t_loc));
      }
  } 
  sum = (-2.0) * sum;
  return(sum);
}



void log_P_alpha_beta(datastr, alpha, beta, res_out)
struct data *datastr;
double *alpha, *beta;
double *res_out;
{
  int *D = datastr->disease_status; /* Disease status */
  double *G = datastr->x; /* Genotypes */
  int *GRP = datastr->grp_id; /* Group number */
  int *Z = datastr->alloc_z; /* Cluster for each group */
  int n = datastr->num_of_subj; /* length of D, G, and GRP */
  int k = datastr->num_of_cluster; /* number of clusters */
  
  double alpha_min = datastr->alpha_min, alpha_max = datastr->alpha_max,
    beta_min = datastr->beta_min,  beta_max = datastr->beta_max;
  double alpha_mu=datastr->alpha_mu, alpha_sigma2=datastr->alpha_sigma2, beta_mu=datastr->beta_mu, beta_sigma2=datastr->beta_sigma2;

  double  *ptrG, sum = 0.0, beta_zi, t_loc, alpha_zi, gi;
  int i,  zi, *ptrD,  *ptrGRP, grpi, method_alpha=datastr->method_alpha, method_beta=datastr->method_beta;

  for (i=0, ptrD=D, ptrG=G, ptrGRP=GRP; i<n; i++, ptrD++, ptrG++, ptrGRP++) {
      grpi = *ptrGRP;
      zi = Z[grpi]; 
      gi = *ptrG; 
      beta_zi = beta[zi]; /* beta remains fixed in the formula */
      alpha_zi = alpha[zi];
      
      t_loc = exp(alpha_zi + beta_zi * gi);
      if (*ptrD) {
         /* for a case */    
        sum += log(t_loc / (1 + t_loc));
      } else {
        /* for a control */
	sum += log(1.0/(1.0 + t_loc));
      }
  } 
  res_out[0] = -2.0 * sum; /*dev */
  /* Now add in log(P(alpha, beta | k)) */ 
  for (i=0; i<k; i++ )
  {
      sum += log(pri_pd_alpha(alpha[i], alpha_min, alpha_max, method_alpha, alpha_mu, alpha_sigma2)) +
             log(pri_pd_beta(beta[i], beta_min, beta_max, method_beta, beta_mu, beta_sigma2));
  }
  res_out[1] = sum;
}

/* using offset */
void log_P_alpha_beta_tao(datastr, alpha, beta, tao, offset, res_out)
struct data *datastr;
double *alpha, *beta, *tao;
double *res_out;
double *offset;
{
  int *D = datastr->disease_status; /* Disease status */
  double *G = datastr->x; /* Genotypes */
  int *GRP = datastr->grp_id; /* Group number */
  int *Z = datastr->alloc_z; /* Cluster for each group */
  int n = datastr->num_of_subj; /* length of D, G, and GRP */
  int k = datastr->num_of_cluster; /* number of clusters */
 
  double alpha_min = datastr->alpha_min, alpha_max = datastr->alpha_max,
    beta_min = datastr->beta_min,  beta_max = datastr->beta_max,
    *tao_min = datastr->tao_min, *tao_max = datastr->tao_max,
    alpha_mu=datastr->alpha_mu, alpha_sigma2=datastr->alpha_sigma2, beta_mu=datastr->beta_mu, beta_sigma2=datastr->beta_sigma2,
    *tao_mu=datastr->tao_mu, *tao_sigma2=datastr->tao_sigma2;

  double  sum = 0.0, beta_zi, t_loc, alpha_zi;
  int i, zi, *ptrD, *ptrGRP, grpi, method_alpha=datastr->method_alpha, method_beta=datastr->method_beta, method_tao=datastr->method_tao ;
  int ncov = datastr->num_of_covars;
  double *ptrG, gi;

  for (i=0, ptrD=D, ptrG=G, ptrGRP=GRP; i<n; i++, ptrD++, ptrG++, ptrGRP++) {
      grpi = *ptrGRP;
      zi = Z[grpi]; 
      gi = *ptrG; 
      beta_zi = beta[zi]; /* beta remains fixed in the formula */
      alpha_zi = alpha[zi];
      t_loc = alpha_zi + beta_zi * gi + offset[i];
      t_loc = exp(t_loc);
      if (*ptrD) {
         /* for a case */    
        sum += log(t_loc / (1 + t_loc));
      } else {
        /* for a control */
	sum += log(1.0 / (1.0 + t_loc));
      }
  } 
  res_out[0] = -2.0 * sum; /*dev */
  /* Now add in log(P(alpha, beta, tao| k)) */ 
  for (i=0; i<k; i++ )
  {
      sum += log(pri_pd_alpha(alpha[i], alpha_min, alpha_max, method_alpha, alpha_mu, alpha_sigma2)) +
             log(pri_pd_beta(beta[i], beta_min, beta_max, method_beta, beta_mu, beta_sigma2));
  }
  if (method_tao == 0) {
    for (i=0; i<ncov; i++) sum += log(pri_pd_tao(tao[i], tao_min[i], tao_max[i], method_tao, 0.0, 0.0));
  } else {
    for (i=0; i<ncov; i++) sum += log(pri_pd_tao(tao[i], 0.0, 0.0, method_tao, tao_mu[i], tao_sigma2[i]));
  }
  
  res_out[1] = sum;
}

/**********************************************************/
/* Functions to update alpha and beta */
/**********************************************************/

/* Function to generate a new vector (f* or beta* in sec 3.1) */
void getVecStar(invec, outvec, k, sigma2)
double *invec, *outvec, sigma2;
int k;
{
  double temp[k], xi;

  /* Generate k N(0, 1) values and store in temp */
  rnormal_vec(0.0, 1.0, temp, k);

  /* Normalize the vector */
  vecnorm(temp, temp, k);

  /* Generate a single N(0, sigma^2) value */
  xi = rnormal(0.0, sigma2);

  /* Compute invec + xi*temp */
  Vec1_p_cVec2(outvec, invec, temp, k, xi);
}

/* Main function to update alpha and beta, returns 1 if updated, 0 otherwise */
int update_alpha_beta(datastr)
struct data *datastr;
{
  double sigma2_alpha = datastr->sigma2_alpha;
  double sigma2_beta = datastr->sigma2_beta;
  int k=datastr->num_of_cluster;
  int i, ret;
  double *ptra1, *ptra2, *ptrb1, *ptrb2, alphastar[datastr->num_of_cluster], betastar[datastr->num_of_cluster], R;
  double res_out_1[2], res_out_2[2];
  double logR;

  /* alpha */
  getVecStar(datastr->alpha, alphastar, k, sigma2_alpha);
  if (datastr->method_alpha == 0) {
    if (checkVecBounds(alphastar, k, datastr->alpha_min, datastr->alpha_max)) return(0);
  }

  /* beta */
  getVecStar(datastr->beta, betastar, k, sigma2_beta);
  if (datastr->method_beta == 0) {
    if (checkVecBounds(betastar, k, datastr->beta_min, datastr->beta_max)) return(0);
  }

  /* Compute R */
  log_P_alpha_beta(datastr,  alphastar, betastar, res_out_1);

  log_P_alpha_beta(datastr,  datastr->alpha, datastr->beta, res_out_2);
      
  
  datastr->dev_i = res_out_2[0]; /* save the dev for the current value */
  logR = res_out_1[1] - res_out_2[1];

  R = logR_to_R(logR);

  /* Update only if returned bernoulli value is 1 */
  ret = bernoulli(R);
  if (ret) {
    for (i=0, ptra1=datastr->alpha, ptra2=alphastar, ptrb1=datastr->beta, ptrb2=betastar; i<k; i++, ptra1++, ptra2++, ptrb1++, ptrb2++) {
      *ptra1 = *ptra2;
      *ptrb1 = *ptrb2;
    }
    
    datastr->dev_i = res_out_1[0];
  }
  return(ret);
}


int update_alpha_beta_tao(datastr)
struct data *datastr;
{
  double sigma2_alpha = datastr->sigma2_alpha;
  double sigma2_beta = datastr->sigma2_beta;
  double sigma2_tao = datastr->sigma2_tao;
  int k=datastr->num_of_cluster;
  int ncov = datastr->num_of_covars;
  int i, j, ret, n=datastr->num_of_subj;
  double *ptra1, *ptra2, *ptrb1, *ptrb2, alphastar[datastr->num_of_cluster], betastar[datastr->num_of_cluster], R;
  double logR;
  double taostar[datastr->num_of_covars];
  double res_out_1[2], res_out_2[2];
  double offset[n];


  /* alpha */
  getVecStar(datastr->alpha, alphastar, k, sigma2_alpha);
  if (datastr->method_alpha == 0) {
    if (checkVecBounds(alphastar, k, datastr->alpha_min, datastr->alpha_max)) return(0);
  }

  /* beta */
  getVecStar(datastr->beta, betastar, k, sigma2_beta);
  if (datastr->method_beta == 0) {
    if (checkVecBounds(betastar, k, datastr->beta_min, datastr->beta_max)) return(0);
  }

  /* tao */
  getVecStar(datastr->tao, taostar, ncov, sigma2_tao);
  if (datastr->method_tao == 0) {
    for (i=0; i<ncov; i++) { 
      if (checkVecBounds(&taostar[i], 1, datastr->tao_min[i], datastr->tao_max[i])) return(0);
    }
  }

 /* Compute R */

  for (i = 0; i < n; i++)
    {
      offset[i] = 0.0;
      for (j = 0; j < ncov; j++)
	{
	  offset[i] = offset[i] + (datastr->covars)[i][j] * taostar[j];
        }
    }
      
 
  log_P_alpha_beta_tao(datastr,  alphastar, betastar, taostar, offset, res_out_1);

  log_P_alpha_beta_tao(datastr,  datastr->alpha, datastr->beta, datastr->tao,
                       datastr->offset, res_out_2);
      
 
  
  datastr->dev_i = res_out_2[0]; /* save the dev for the current value */

  /* no need to update offset if tao is not updated */  
  logR = res_out_1[1] - res_out_2[1];
  R = logR_to_R(logR);
  
  /* Update only if returned bernoulli value is 1 */
  ret = bernoulli(R);
  if (ret) {
    for (i=0, ptra1=datastr->alpha, ptra2=alphastar, ptrb1=datastr->beta, ptrb2=betastar; i<k; i++, ptra1++, ptra2++, ptrb1++, ptrb2++) {
      *ptra1 = *ptra2;
      *ptrb1 = *ptrb2;
    }
    
    for (i = 0, ptra1=datastr->tao, ptra2 = taostar; i < ncov; i++, ptra1++, ptra2++)
      {
        *ptra1 = *ptra2;
      }       
  
    for (i = 0, ptra1=datastr->offset, ptra2 = offset; i < n; i++, ptra1++, ptra2++)
      {
        *ptra1 = *ptra2;
      } 
    datastr->dev_i = res_out_1[0];
  }
  return(ret);
}

/**********************************************************/
/* Functions to update z */
/**********************************************************/

/* Function to compute sum of similarity scores for subjects in same cluster */
double compute_tij(sim_row, z, n, j, neighbor_id, num_neighbor)
double *sim_row;  /* A single row of the similarity matrix */
int *z; /* cluster allocation vector with values 0, ..., k-1 */
int n; /* number of units (length of sim_row */
int j; /* Cluster number */
int *neighbor_id, num_neighbor;
{
  int i, neigh;
  double sum = 0.0;

  if (!num_neighbor) return(sum);
  for (i=0; i<num_neighbor; i++) {
    neigh = neighbor_id[i];
    if (z[neigh] == j) sum += sim_row[neigh];
  }

  return(sum);
}

void compute_ti(outvec, sim_row, z, n, k, neighbor_id, num_neighbor)
double *outvec; /* output vector of similarity sums for each cluster (length k) */
double *sim_row;  /* A single row of the similarity matrix */
int *z; /* cluster allocation vector with values 0, ..., k-1 */
int n; /* number of subjects (length of sim_row) */
int k; /* number of clusters */
int *neighbor_id, num_neighbor;
{
  int i, neigh;

  /* Initialize vector to 0 */
  vecinit(outvec, k, 0.0);
 
  for (i=0; i<num_neighbor; i++) {
    neigh = neighbor_id[i];
    outvec[z[neigh]] += sim_row[neigh];
  }  

}

/* Function to compute the log_probability P(z = j| ...) for a single group */
double P_z_j(d, g, offset, subj_list, n_subj, alpha, beta, psi, tij)
int *d;  /* disease status */
double *g; /* genotype 0-1-2 */
double *offset; /*from the covariates, which won't change by z */

int *subj_list; /*list of subject in this group */
int n_subj; /* the number of subjects in this group */
double alpha, beta, psi;
double tij; /* Sum of similarity scores */
{
  int jj, i_subj;
  double t_1, t_2;
  t_1 = psi*tij;

  for (jj=0; jj < n_subj; jj++) {
      i_subj = subj_list[jj];
      t_2 = exp(alpha + beta * g[i_subj] + offset[i_subj]);
      if (d[i_subj] == 1) {
          /* for a case */
        t_1 =  t_1 + log( t_2 / (1.0 + t_2));
      } else {
	t_1 = t_1 - log(1.0 + t_2);
      }
  }
 
  return(t_1);
}

int accept_or_not_grp(n_grp_cluster, k, cluster_size)
int *n_grp_cluster;
int k;
double cluster_size;
{
  int i, j;
  j = 1;
  for (i = 0; i < k; i++)
    {
      if (n_grp_cluster[i] <= cluster_size)
	{
	  j = 0;
	  break;
	}
    }
  return(j);
}

int accept_or_not(v_case_cluster, v_control_cluster, k, prop_cut, cluster_size)
int *v_case_cluster;
int *v_control_cluster;
int k;
double prop_cut;
double cluster_size;
{
  int i, j;
  int v_max, v_min;
  double r_ck, r_tot;
 
  j = 1;
  for (i=0; i < k; i++)
    {
      v_max = MAX(v_case_cluster[k], v_control_cluster[k]);
      v_min = MIN(v_case_cluster[k], v_control_cluster[k]);
      
      if (v_max >0)
	{
          r_tot = v_max + v_min;
	  r_ck = v_min/ r_tot;
          if (r_ck <= prop_cut || r_tot <= cluster_size)
	    {
	      j = 0;
	      break;
	    }
	} 
    }
  return(j);    
}



/* Function to obtain the cluster counts with weight  */
void cluster_counts_weight(outvec, k, z, ngrp, vec_weight)
int *outvec;  /* length k */
int k;
int *z;  /* Must be labeled 0, 1, ..., k-1 */
int ngrp; /* length of z */
int *vec_weight; /* weight for each postion at z  */
{
  int i;

  /* Initialize outvec to 0 */
  ivecinit(outvec, k, 0);

  for (i=0; i<ngrp; i++) 
    {  
      outvec[z[i]] = outvec[z[i]] + vec_weight[i];
    }
  
}


/* Function to obtain the cluster counts.  */
void cluster_counts_case(outvec, k, z, n, y, id_y)
int *outvec;  /* length k */
int k;
int *z;  /* Must be labeled 0, 1, ..., k-1 */
int n; /* length of z */
int *y;  
int id_y;
{
  int i, *ptr;

  /* Initialize outvec to 0 */
  ivecinit(outvec, k, 0);

  for (i=0, ptr=z; i<n; i++, ptr++) 
    {  
       if (y[i] == id_y)
        {
          outvec[*ptr]++;
        }
    } 
}

/* Function to obtain the cluster counts.  */
void cluster_counts(outvec, k, z, n)
int *outvec;  /* length k */
int k;
int *z;  /* Must be labeled 0, 1, ..., k-1 */
int n; /* length of z */
{
  int i, *ptr;

  /* Initialize outvec to 0 */
  ivecinit(outvec, k, 0);

  for (i=0, ptr=z; i<n; i++, ptr++) outvec[*ptr]++; 
}

/* translate logR into R with R from 0 to 1 */
double logR_to_R(logR)
     double logR;
{
  double R;
  if (logR >= 0.0){
    R = 1.0;
  } else if (logR < -30) {
    R = 0.0;
  } else {
    R = exp(logR);
  }
  return(R);
}

  
/* Function to update z on the space of subgroup*/
int update_z(datastr)
struct data *datastr;
{
  int i,  k = datastr->num_of_cluster, *z = datastr->alloc_z;
  int ngrp = datastr->num_of_grp;
  int ntimes, sub, j, jstar, *D=datastr->disease_status;
  double  *G=datastr->x;

  int n_type1[datastr->num_of_cluster];
  int id_accept;

  double tij, tijstar, Pj, Pjstar, *alpha=datastr->alpha, *beta=datastr->beta, 
         **sim_mat=datastr->simi_mat, *sim_row;
  double psi=datastr->psi_interaction, prob, logR;
  double update_prop=datastr->update_prop;
  double *offset = datastr->offset;
  int **subj_grp = datastr->subj_grp, *n_subj_grp = datastr->n_subj_grp, ngrpm1;
 
  int *neigh_id, num_neigh, **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor;

  /* Return if only 1 cluster */
  if (k == 1) return(0);

  ntimes = floor(ngrp*update_prop);
  if (ntimes <= 0) return(0);
  if (update_prop > 0.99)
    {
      /* if rate is 1, go through each element */
      ntimes = ngrp;
    }

  cluster_counts(n_type1, k, z, ngrp); 

  ngrpm1 = ngrp - 1;
  for (i=0; i<ntimes; i++) {
    if (update_prop > 0.99)
      {
	sub = i;
      } 
    else
      {
	/*sub = gsl_rng_uniform_int(rngptr, ngrp);*/
       sub = uniform_int(0, ngrpm1);
      }

    j = z[sub];
  
    /* Randomly pick a cluster other than j */
    /* jstar < j, don't change, but when >=j, shift one */
    /*jstar = gsl_rng_uniform_int(rngptr, k-1);*/
    jstar = uniform_int(0, k-2);

    if (jstar >= j)
    {
		  jstar=jstar + 1;
    }

    n_type1[jstar]++;
    n_type1[j]--;
  
    /* make sure it is elgible */
    id_accept = accept_or_not_grp(n_type1, k, datastr->cluster_size);

    if (id_accept != 1)
      {
        /* not accept */
       
	n_type1[jstar]--;
	n_type1[j]++;
	continue;
      }           

  
    /* Calculate P(z_i = j) and P(z_i = jstar) */
  
    sim_row = sim_mat[sub];
    neigh_id = neighbor_id[sub];
    num_neigh = num_neighbor[sub]; 
    tij = compute_tij(sim_row, z, ngrp, j, neigh_id, num_neigh);
    tijstar = compute_tij(sim_row, z, ngrp, jstar, neigh_id, num_neigh);
    Pj = P_z_j(D, G, offset,  subj_grp[sub], n_subj_grp[sub], alpha[j], beta[j], psi, tij);
    Pjstar = P_z_j(D, G, offset, subj_grp[sub], n_subj_grp[sub], alpha[jstar], beta[jstar], 
                   psi, tijstar);

 
    /* Get a bernoulli random value to update or not */
    
    logR = Pjstar - Pj;
    prob = logR_to_R(logR);

    /*    prob = exp(Pjstar - Pj)*n_1/n_2; */
  
    if (bernoulli(prob)) {
      /* Update */
      z[sub] = jstar;      
    } else {
   	n_type1[jstar]--;
	n_type1[j]++;
    }
  }

  /* don't need to check as we make sure the same cluster number */
  /* check_z(datastr); */
  return(1);
}


/* Function to update z on the space of subgroup, based on w0*/
int update_1w(datastr, w0, z)
struct data *datastr;
int *w0;
int *z;
{
  int i,  k = datastr->num_of_cluster;
  int ngrp = datastr->num_of_grp;
  int ntimes,  sub, j, jstar, ngrpm1;
 
  int id_accept;
  int n_type1[datastr->num_of_cluster];

  double tij, tijstar,
         **sim_mat=datastr->simi_mat, *sim_row;
  double psi=datastr->psi_interaction, prob;
  double update_prop=datastr->update_prop_w;
  int *neigh_id, num_neigh, **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor;

  ivec_ivec_init(z, w0, ngrp);

  /* Get the cluster counts */
  cluster_counts(n_type1, k, z, ngrp); 

  /* Return if only 1 cluster */
  if (k == 1) return(0);

  ntimes = floor(ngrp*update_prop);
  if (ntimes <= 0) return(0);
  if (update_prop > 0.99)
    {
      /* if rate is 1, go through each element */
      ntimes = ngrp;
    }

  ngrpm1 = ngrp - 1;
  for (i=0; i<ntimes; i++) {
    if (update_prop > 0.99)
      {
	sub = i;
      } 
    else
      {
	/*sub = gsl_rng_uniform_int(rngptr, ngrp);*/
       sub = uniform_int(0, ngrpm1);
      }

    j = z[sub];
  
    /* Randomly pick a cluster other than j */
    /* jstar < j, don't change, but when >=j, shift one */
    /*jstar = gsl_rng_uniform_int(rngptr, k-1);*/
    jstar = uniform_int(0, k-2);
    if (jstar >= j)
    {
      jstar=jstar + 1;
    }
    
    n_type1[jstar]++;
    n_type1[j]--;
  
    /* make sure it is elgible */
    id_accept = accept_or_not_grp(n_type1, k, datastr->cluster_size);

    if (id_accept != 1)
      {
        /* not accept */
       
	n_type1[jstar]--;
	n_type1[j]++;
	continue;
      }       

    /* Calculate P(z_i = j) and P(z_i = jstar) */
  
    sim_row = sim_mat[sub];
    neigh_id = neighbor_id[sub];
    num_neigh = num_neighbor[sub]; 
    tij = compute_tij(sim_row, z, ngrp, j, neigh_id, num_neigh);
    tijstar = compute_tij(sim_row, z, ngrp, jstar, neigh_id, num_neigh);
    prob = exp(psi * (tijstar-tij));
   
    /*    prob = exp(Pjstar - Pj)*n_1/n_2; */
  
    if (bernoulli(prob)) {
      /* Update */
      z[sub] = jstar;      
    } else {
   	n_type1[jstar]--;
	n_type1[j]++;
      
    }
  }

  return(1);
}

int update_all_w(datastr, w0, W)
struct data *datastr;
int *w0;
int **W;
{
   int m=datastr->w_m;
   int i;
   for (i = 0; i < m; i ++)
     {
       if (!i)
	 {
           update_1w(datastr, w0, W[i]);
	 } else {
	 update_1w(datastr, W[i-1], W[i]);
       }
     }
  return(0);
}
   

/***********************************************************/
/* Functions for updating the interaction parameter */
/***********************************************************/


/* Function for generating 1 w vector */
void gen_alloc_cond1(outvec, psi, k, sim_mat, w0, n, neighbor_id, num_neighbor)
int *outvec;
double psi, **sim_mat;
int k, n;
int *w0; /* Initial allocation vector */ 
int **neighbor_id, *num_neighbor;
{
  double prob[k], ti[k], *ptrd, *ptrti;
  int i, j, *ptrout;  

  /* Initialize outvec to w0 */
  ivec_ivec_init(outvec, w0, n);
 
  /* Loop over each subject */
  for (i=0, ptrout=outvec; i<n; i++, ptrout++) {

    /* Compute tij for j = 0, ..., k-1 */
    compute_ti(ti, sim_mat[i], outvec, n, k, neighbor_id[i], num_neighbor[i]);

    /* Compute probabilities P ~~ exp(psi*tij), j = 0, ..., k-1 */
    for (j=0, ptrd=prob, ptrti=ti; j<k; j++, ptrd++, ptrti++) *ptrd = exp(psi* *ptrti);

    /* Normalize probabilities */
    vecnormSum(prob, prob, k);

    /* Randomly assign new cluster based on the probabilities */
    *ptrout = sample1(prob, k);
  }

}

/* Function for generating m w vectors */
void gen_alloc_cond(outmat, psi, k, sim_mat, w0, n, m, neighbor_id, num_neighbor)
int **outmat; /* Must be of dimension m by n (m rows, n columns) */
double psi, **sim_mat;
int k, n;
int *w0; /* allocation vector */
int m; /* Number of w vectors */ 
int **neighbor_id, *num_neighbor;
{
 
  int i;  

  /* Loop over each vector */
  for (i=0; i<m; i++) {
    /* Generate vector, pass in the previous vector each time */
    if (!i) {
      gen_alloc_cond1(outmat[i], psi, k, sim_mat, w0, n, neighbor_id, num_neighbor);
    } else {
      gen_alloc_cond1(outmat[i], psi, k, sim_mat, outmat[i-1], n, neighbor_id, num_neighbor);
    }
  }
}

/* Function for discrete psi */
int update_psi_discrete(datastr)
struct data *datastr;
{
  int i_psi=datastr->psi_position, i_psi_new, ngrid=datastr->n_psi_grid, ngridm1, psi_bound, psi_new_bound, ret=0;;
  double *grid=datastr->psi_grid, *norm=datastr->normalizing_psi_grid, *prob=datastr->prob_psi_grid;
  double psi, psi_new, p_psi, p_psi_new, u, value, logr, r;

  ngridm1 = ngrid - 1;
  if (i_psi == 0) {
    i_psi_new = 1; 
    psi_bound = 1;
  } else if (i_psi == ngridm1) {
    i_psi_new = ngrid - 2;
    psi_bound = 1;
  } else {
    if (bernoulli(0.5)) {
      i_psi_new = i_psi + 1;
    } else {
      i_psi_new = i_psi - 1;
    }
    psi_bound = 0;
  }
  if ((i_psi_new == 0) || (i_psi_new == ngridm1)) {
    psi_new_bound = 1;
  } else {
    psi_new_bound = 0;
  }

  psi        = grid[i_psi];
  psi_new    = grid[i_psi_new];
  p_psi      = prob[i_psi];
  p_psi_new  = prob[i_psi_new];
  u = calc_Uz(datastr->simi_mat, datastr->alloc_z, datastr->num_of_grp, datastr->neighbor_id, datastr->num_neighbor);
  value = norm[i_psi] - norm[i_psi_new] + log(p_psi_new) - log(psi) + (psi_new - psi)*u;

  if ((!psi_bound) && (!psi_new_bound)) {
    logr = value;
  } else if ((psi_bound) && (!psi_new_bound)) {
    logr = value + LOG_POINT5;
  } else {
    logr = value - LOG_POINT5;
  }

  r = logR_to_R(logr);
  if (bernoulli(r)) {
    ret = 1;

    /* Update */
    datastr->psi_interaction = psi_new;
    datastr->psi_position    = i_psi_new; 
  }

  return(ret);

} /* END: update_psi_discrete */

/* Main function for updating psi */
int update_psi(datastr)
struct data *datastr;
{
  double sigma2_psi=datastr->sigma2_psi;
  int m=datastr->w_m;
  double psi_new, psi=datastr->psi_interaction, psi_diff, **sim_mat=datastr->simi_mat, Rhat=0.0, u, r;
  double p_psi=0.0, p_psi_new,  weight[m], temp, *ptrd;
  int i, n=datastr->num_of_grp, k=datastr->num_of_cluster, **W=datastr->auxillary_sample, *Z=datastr->alloc_z;
  int w0[datastr->num_of_grp], ret=0;
  int **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor;

  /* Return if only 1 cluster */
  if (k == 1) return(0);

  /* Generate a new psi */
  psi_new = psi + rnormal(0.0, sigma2_psi);

  /* Check the new psi */
  if (datastr->method_psi == 0)
    {
      if ((psi_new < datastr->psi_min) || (psi_new > datastr->psi_max)) return(0);
    }
    
  /* Compute R(psi, psi_new) */
  psi_diff = psi_new - psi;
  for (i=0, ptrd=weight; i<m; i++, ptrd++) {
    
    /* Calculate U(w) */
    u = calc_Uz(sim_mat, W[i], n, neighbor_id, num_neighbor);

    temp = exp(psi_diff*u);
    *ptrd = temp;    
    Rhat += temp;
  }
  Rhat /= m;

  /* Compute [p(psi`)exp((psi`-psi)U(z))]/[Rhat*p(psi)]  */
  /*  p_psi = pri_pd_psi_gamma(psi, datastr->psi_k, datastr->psi_theta);*/
  /*     p_psi_new = pri_pd_psi_gamma(psi_new, datastr->psi_k, datastr->psi_theta);*/

  if (datastr->method_psi == 0)
    {
      p_psi = pri_pd_psi_unif(psi, datastr->psi_min, datastr->psi_max);
      p_psi_new = pri_pd_psi_unif(psi_new, datastr->psi_min, datastr->psi_max);
    }

  if (datastr->method_psi == 1)
    {
      p_psi = pri_pd_psi_gamma(psi, datastr->psi_k, datastr->psi_theta);
      p_psi_new = pri_pd_psi_gamma(psi_new, datastr->psi_k, datastr->psi_theta);
    }

  u = calc_Uz(sim_mat, Z, n, neighbor_id, num_neighbor);

  r = (p_psi_new*exp(psi_diff*u))/(Rhat*p_psi);

  if (bernoulli(r)) {
    ret = 1;

    /* Update */
    datastr->psi_interaction = psi_new;

    /* Update auxillary sample. First normalize weights */
    vecnormSum(weight, weight, m);

    /* Generate a random int from 0 to m-1 with probs in weight vector */
    i = sample1(weight, m);

    /* Copy row of W to w0 */
    ivec_ivec_init(w0, W[i], n);

    /* Generate the new set of vectors */
    update_all_w(datastr, w0, W);

  }

  return(ret);
}



/**********************************************************************/
/* Function for checking the allocation vector is from 0, 1, ..., k-1 */
/**********************************************************************/
void check_z(datastr)
struct data *datastr;
{
  int i, j, flag, newk, counts[datastr->num_of_cluster], subtract[datastr->num_of_cluster];
  int k=datastr->num_of_cluster, n=datastr->num_of_grp, *z=datastr->alloc_z, *ptri;
  double *alpha, *beta;

  /* Get the cluster counts */
  cluster_counts(counts, k, z, n);

  /* Determine if any counts are 0 */
  flag = 0;
  for (i=0, ptri=counts; i<k; i++, ptri++) {
    if (*ptri == 0) {
      flag = 1;
      break;
    }
  }

  if (!flag) return;

  /* Initialize subtract vector to 0 */
  ivecinit(subtract, k, 0);
  newk = k;
  alpha = datastr->alpha;
  beta = datastr->beta;

  /* If cluster i has 0 subjects, then subtract 1 from clusters i+1, ..., k */
  /* substract tell the number of shifts to use */
  for (i=0, ptri=counts; i<k; i++, ptri++) {
    if (*ptri == 0) {
      newk--;
      for (j=i+1; j<k; j++) {
        subtract[j]++;
      
        /* Move alpha and beta */
        alpha[j-1] = alpha[j];
        beta[j-1] = beta[j];
      }
    }
  }

  /* Update z and k */
  datastr->num_of_cluster = newk;
  for (i=0, ptri=z; i<n; i++, ptri++) {
    j = subtract[*ptri];
    if (j) *ptri -= j;
  }

}


/* Function to compute a fixed probability for 1 group */
double fixed_prob(datastr, alphajm, alphajp, betajm, betajp, grp_id)
struct data *datastr;
double  betajm, betajp, alphajm, alphajp; /* beta j minus and beta j plus */
int grp_id; /* subgroup id for use */
 {
  int *subj_list, n_subj, jj, i_subj;
  double *g, *offset;
  int *d;
  double t_1, t_2, t_sum_1, t_sum_2, ret;

  subj_list = (datastr->subj_grp)[grp_id];
  n_subj = (datastr->n_subj_grp)[grp_id];
  g = datastr->x;
  d = datastr->disease_status;
  offset = datastr->offset;
 
  t_sum_1 = 0.0;
  t_sum_2 = 0.0;
  for (jj = 0; jj < n_subj; jj++)
    {
      i_subj = subj_list[jj];
      t_1 = exp(alphajm + betajm * g[i_subj] + offset[i_subj]);
      t_2 = exp(alphajp + betajp * g[i_subj] + offset[i_subj]);
      if (d[i_subj] == 1)
	{
	  t_sum_1 = t_sum_1 + log (t_1 / (1.0 + t_1));
          t_sum_2 = t_sum_2 + log (t_2 / (1.0 + t_2));
        } else {
	  t_sum_1 = t_sum_1 - log(1.0 + t_1);
          t_sum_2 = t_sum_2 - log(1.0 + t_2);
      }
    }
  /*  ret = 1.0 / (1.0 + exp(t_sum_2 - t_sum_1));*/
  ret = t_sum_2 - t_sum_1;
  return(ret);
 }



/* Function to compute an adaptive probability for one group , only return t_sum_2 - t_sum_1*/
double adapt_prob(datastr, alphajm, alphajp, betajm, betajp, grp_id, z, jm, jp)
struct data *datastr;
double  betajm, betajp, alphajm, alphajp; /* beta j minus and beta j plus */
int grp_id; /* subgroup id for use */
int *z, jm, jp; /* Cluster numbers for computing tij */
{
  int *subj_list, n_subj, jj, i_subj;
  double *g, *offset;
  int *d, n;
  double *sim_row;
  double psi;
  double t_1, t_2, t_sum_1, t_sum_2, ret, tm, tp;
  int *neigh_id, num_neigh;

  n = datastr->num_of_grp;
  subj_list = (datastr->subj_grp)[grp_id];
  n_subj = (datastr->n_subj_grp)[grp_id];
  g = datastr->x;
  d = datastr->disease_status;
  offset = datastr->offset;
  sim_row = datastr->simi_mat[grp_id];
  psi = datastr->psi_interaction;
  neigh_id = datastr->neighbor_id[grp_id];
  num_neigh = datastr->num_neighbor[grp_id];

  tm = compute_tij(sim_row, z, n, jm, neigh_id, num_neigh);
  tp = compute_tij(sim_row, z, n, jp, neigh_id, num_neigh);

  t_sum_1 = psi * tm;
  t_sum_2 = psi * tp;
  for (jj = 0; jj < n_subj; jj++)
    {
      i_subj = subj_list[jj];
      t_1 = exp(alphajm + betajm * g[i_subj] + offset[i_subj]);
      t_2 = exp(alphajp + betajp * g[i_subj] + offset[i_subj]);
      if (d[i_subj] == 1)
	{
	  t_sum_1 = t_sum_1 + log (t_1 / (1.0 + t_1));
          t_sum_2 = t_sum_2 + log (t_2 / (1.0 + t_2));
        } else {
	  t_sum_1 = t_sum_1 - log(1.0 + t_1);
          t_sum_2 = t_sum_2 - log(1.0 + t_2);
      }
    }
  /* ret = 1.0 / (1.0 + exp(t_sum_2 - t_sum_1)); */
  ret = t_sum_2 - t_sum_1;
  return(ret);
}

/* the prob is either as p=1/(1+exp(t_diff)) if id_prob=0,  or 1-p if id_prob = 1 */
double logprob_calculate(t_diff, id_prob)
     double t_diff;
     int id_prob;
{
  double logp;
  double BIG_VALUE = 30.0;
  double TINY_VALUE = -30.0;

  if (id_prob == 0)
    {
      if(t_diff > BIG_VALUE){ 
	logp = t_diff * (-1.0);
      } else if (t_diff < TINY_VALUE) 
	{
	  logp = 0.0;
	} else {
	logp = log(1+exp(t_diff)) * (-1.0);
      }
    } else {
    if(t_diff > BIG_VALUE){ 
      logp = 0.0;
    } else if (t_diff < TINY_VALUE) 
      {
	logp = t_diff;
      } else {
      logp = t_diff - log(1+exp(t_diff));
    }
  }	
  
  return(logp);
}



/* function to split cluster jstar into jm and jp, and new assignment is at newZ */
/* the newZ is the same as Z at the begining, but being updated at each step */

double assign_split(datastr, alphajm, alphajp,  betajm, betajp, newZ, jstar, jm, jp)
     struct data *datastr;
     double alphajm, alphajp, betajm, betajp;
     int *newZ, jstar, jm, jp;
{
  int *Z; 
  int method;
  double p_alloc = 0.0, t1, prob;
  int i, n;
 
  Z = datastr->alloc_z;
  method = datastr->split_option;
  n = datastr->num_of_grp;

  for (i = 0; i < n; i++)
    {
      if (Z[i] == jstar)
	{
	  if (method == 0)
	    {
	      t1 = adapt_prob(datastr, alphajm, alphajp, betajm, betajp, i, newZ, jm, jp);
	    } else {
	    t1 = fixed_prob(datastr,  alphajm, alphajp, betajm, betajp, i);
	  }

	  if(t1 > 30){ 
	    prob = 0.0;
	  } else if (t1 < -30) {
	    prob = 1.0;
	  } else {
	    prob = 1.0 / (1.0 + exp(t1));
	  }

	  if (bernoulli(prob)) {
	    newZ[i] = jm;
	    p_alloc = p_alloc + logprob_calculate(t1, 0);
	  } else {
	    newZ[i] = jp;
	    p_alloc = p_alloc + logprob_calculate(t1, 1); 
	  }
	}
    }
  return(p_alloc);
}
 
/* Function used for merging to assign new subjects (method = adapt) and return log(p_alloc) */
/* Z is the original allocation vector, newZ is new merged allocation vector, but will not be updated */
/* f beta are for the value before merge */

double assign_merge(datastr, alphajm, alphajp,  betajm, betajp, newZ, jstar, jm, jp)
     struct data *datastr;
     double alphajm, alphajp, betajm, betajp;
     int *newZ, jstar, jm, jp;
{
  int *Z; 
  int method, i;
  double p_alloc = 0.0, t1;
  int  n = datastr->num_of_grp;
  int Z_update[n];
  
 
  Z = datastr->alloc_z;
  method = datastr->split_option;
  
  ivec_ivec_init(Z_update, newZ, n);
  
  /* newZ is the merged one, Z is the original one, Z_update is the one keep updating */
  for (i = 0; i < n; i++)
    {
      if (newZ[i] == jstar)
	{
	  if (method == 0)
	    {
	      t1 = adapt_prob(datastr, alphajm, alphajp, betajm, betajp, i, Z_update, jm, jp);
	    } else {
	    t1 = fixed_prob(datastr,  alphajm, alphajp, betajm, betajp, i);
	  }
          Z_update[i] = Z[i];
	
          if (Z[i] == jm)
	    {
	      p_alloc = p_alloc + logprob_calculate(t1, 0);
	    } else {
	    p_alloc = p_alloc + logprob_calculate(t1, 1);
	  }
	  
	}

    }
  return(p_alloc);
}


/* Function to search a vector of ranks for a particular rank and return the index */
int rank_index(rank_vec, n, rank)
int *rank_vec, rank, n;
{
  int *ptr, i;

  for (i=0, ptr=rank_vec; i<n; i++, ptr++) {
    if (*ptr == rank) return(i);
  }
  return(-9999);
}

/* Function to compute the sum of log probabilities needed in calculation of R for a specific cluster */
double log_prob_1clus(datastr, alpha, beta, z, cluster)
     struct data *datastr;
     double alpha, beta;
     int *z, cluster;
{
  double *g, *offset;
  int *d, *GRP;
  double sum, t_1;
  int num_of_subj, i, grpi;

  GRP = datastr->grp_id;
  g = datastr->x;
  d = datastr->disease_status;
  offset = datastr->offset;
  num_of_subj = datastr->num_of_subj;

  sum=0.0;

  for (i = 0; i < num_of_subj; i++) 
    {
      grpi = GRP[i];
      if (z[grpi] == cluster)
	{
	  t_1 = exp(alpha + beta * g[i] + offset[i]);
	  if (d[i] == 1)
	    {
	      sum = sum + log (t_1 / (1.0 + t_1));
	    } else {
	    sum = sum - log(1.0 + t_1);
	  }
	}
    }
  return(sum);
}

/* Function to return beta_j1* and betaj2* for the computation of R */
void get_betaj12_merge(beta, rank, k, rankjminus, rankjplus, betaj1, betaj2, BETA_MIN, BETA_MAX)
double *beta; /* Vector of beta */
int *rank; /* Rank of each beta */
int k; /* Length of beta, rank (before merge) */
int rankjminus, rankjplus;
double *betaj1, *betaj2; /* return values */
double BETA_MIN, BETA_MAX;
{
  int i, pos1=-1, pos2=-1;

  if (k == 2) {
    /* Only 2 clusters before merge */
    *betaj1 = BETA_MIN;
    *betaj2 = BETA_MAX;
    return;
  }

  if (rankjminus == 0) {
    /* Left endpoint */
    *betaj1 = BETA_MIN;
    pos2    = 2;
  } else if (rankjplus == k-1) {
    /* Right endpoint */
    *betaj2 = BETA_MAX;
    pos1    = k - 3;
  } else {
    pos1 = rankjminus - 1;
    pos2 = rankjplus + 1;
  }

  if (pos1 > -1) {
    i = rank_index(rank, k, pos1);
    *betaj1 = beta[i];
  }
  if (pos2 > -1) {
    i = rank_index(rank, k, pos2);
    *betaj2 = beta[i];
  }

}



int fun_merge(datastr)
struct data *datastr;
{
  int k=datastr->num_of_cluster, pos, pos1, pos2, jstar, jminus=-1, jplus=-1, i;
  int *ptri, *ptri2, *z=datastr->alloc_z, zprime[datastr->num_of_grp], n=datastr->num_of_grp, counts[datastr->num_of_cluster+1];
  int k0, ret=0, order[datastr->num_of_cluster];
  double *beta=datastr->beta, alphajminus, alphajplus, betajminus, betajplus, alphajstar, betajstar;
  double *alpha=datastr->alpha, R, R1, R2, R3, R4, R5, **sim_mat=datastr->simi_mat, num, den, logp_alloc;
  double betaj1, betaj2;
  double psi=datastr->psi_interaction;
  double shape_split_a, shape_split_b, shape_merge_a, shape_merge_b;
  int **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor;

  if (k < 2) return(0);
  jstar = k; /* Cluster numbers go from 0 to k for the new allocation, but two components are empty*/

  
  get_ranks(beta, k, order);

  /* Generate a random integer from 0 to k-1 */

  /* pos1 and pos2 are the ranking, jminus and jplis are index for cluster */
  /*pos1 = gsl_rng_uniform_int(rngptr, k-1);*/
  pos1 = uniform_int(0, k-2);
  pos2 = pos1 + 1;
  jminus = rank_index(order, k, pos1);
  jplus = rank_index(order, k, pos2);

  /* get betaj1* and betaj2* */
  get_betaj12_merge(beta, order, k, pos1, pos2, &betaj1, &betaj2, datastr->beta_min, datastr->beta_max);

  /* Get f-, beta-, f+, beta+ */
  alphajminus = alpha[jminus];
  alphajplus  = alpha[jplus];
  betajminus = beta[jminus];
  betajplus  = beta[jplus];

  /* Generate f* and beta* for the new cluster */
  /* alphajstar = uniform(rngptr, MIN(alphajminus, alphajplus), MAX(alphajminus, alphajplus));*/
  /*betajstar = uniform(rngptr, betajminus, betajplus); */
 
  shape_split_a = datastr->shape_split_a;
  shape_split_b = datastr->shape_split_b;
  shape_merge_a = datastr->shape_merge_a;
  shape_merge_b = datastr->shape_merge_b;

  alphajstar = gen_beta(alphajminus, alphajplus, shape_merge_a, shape_merge_b);
  betajstar = gen_beta(betajminus, betajplus, shape_merge_a, shape_merge_b);


   /* Get the new allocation vector, zprime is not changed in log_p_alloc */
  for (i=0, ptri=z, ptri2=zprime; i<n; i++, ptri++, ptri2++) {
    pos = *ptri;
    if ((pos == jminus) || (pos == jplus)) {
      *ptri2 = jstar;
    } else {
      *ptri2 = pos;
    }
  }

   /* when do merge, the zprime is not changed */
  logp_alloc = assign_merge(datastr, alphajminus, alphajplus, betajminus, betajplus,
			    zprime, jstar, jminus, jplus);

  /* Get k0 = the number of new clusters with more than 1 subject */
  cluster_counts(counts, k+1, zprime, n);
  k0 = 0;
  for (i=0, ptri=counts; i<=k; i++, ptri++) {
    if (*ptri >= (2* datastr->cluster_size)) k0++;
  }

  /* Calculate components of R (on log scale) */
  R1 = log_prob_1clus(datastr, alphajstar, betajstar, zprime, jstar) - 
    log_prob_1clus(datastr, alphajminus, betajminus, z, jminus) - 
    log_prob_1clus(datastr, alphajplus, betajplus, z, jplus);


  R2 = log(pri_pd_alpha(alphajstar, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2)) +
	   log(pri_pd_beta(betajstar, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2)) -
	   log(pri_pd_alpha(alphajminus, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2)) -
       log(pri_pd_alpha(alphajplus, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2))  -
       log(pri_pd_beta(betajminus, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2)) -
       log(pri_pd_beta(betajplus, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2));

  R3 = psi*(calc_Uz(sim_mat, zprime, n, neighbor_id, num_neighbor) - calc_Uz(sim_mat, z, n, neighbor_id, num_neighbor));

  num =  prob_beta(MIN(alphajminus, alphajplus), datastr->alpha_min, alphajstar, shape_split_b, shape_split_a) *
    prob_beta(MAX(alphajminus, alphajplus), alphajstar, datastr->alpha_max, shape_split_a, shape_split_b) *
    prob_beta(betajminus, betaj1, betajstar, shape_split_b, shape_split_a) *
    prob_beta(betajplus, betajstar, betaj2,  shape_split_a, shape_split_b) /
    prob_beta(alphajstar, alphajminus, alphajplus, shape_merge_a, shape_merge_b) /
    prob_beta(betajstar, betajminus, betajplus, shape_merge_a, shape_merge_b) / 2.0;

  den = (datastr->b[k-2])*(k-1) / (k0*(datastr->d[k-1]));
  R4 = log(num * den) + logp_alloc;

  /* note, b[k-1] is associated with k cluster */
  /* num = (datastr->b[k-2])*(k-1)*fabs(alphajplus-alphajminus)*(betajplus-betajminus);*/
  /*  den = k0*(datastr->d[k-1])*(alphajstar-datastr->alpha_min)*(datastr->alpha_max-alphajstar)* */
  /* (betajstar-betaj1)*(betaj2-betajstar); */

  /*R4 = logp_alloc + log(num/den);*/

  R5 = R1 + R2 + R3 + R4;
  R = logR_to_R(R5);




  if (bernoulli(R)) {
    /* Accept the merge */
    ivec_ivec_init(z, zprime, n);
    alpha[jstar] = alphajstar;
    beta[jstar] = betajstar;

    /* Set num of clusters to k + 1, and let the check_z function make the correction */
    datastr->num_of_cluster = k + 1;
    check_z(datastr);
    ret = 1;
  }
  return(ret);
}


/************************************************************************************************/
/*  Function to split a cluster */
/************************************************************************************************/

/* Function to return beta_j1 and betaj2 for the computation of R */
void get_betaj12_split(beta, rank, k, rankjstar, betaj1, betaj2, BETA_MIN, BETA_MAX)
double *beta; /* Vector of beta */
int *rank; /* Rank of each beta */
int k; /* Length of beta, rank (before merge) */
int rankjstar;
double *betaj1, *betaj2; /* return values */
double BETA_MIN, BETA_MAX;
{
  int i, pos1=-1, pos2=-1;

  if (k == 1) {
    /* Only 1 cluster before split */
    *betaj1 = BETA_MIN;
    *betaj2 = BETA_MAX;
    return;
  }

  if (rankjstar == 0) {
    /* Left endpoint */
    *betaj1 = BETA_MIN;
    pos2    = 1;
  } else if (rankjstar == k-1) {
    /* Right endpoint */
    *betaj2 = BETA_MAX;
    pos1    = k - 2;
  } else {
    pos1 = rankjstar - 1;
    pos2 = rankjstar + 1;
  }

  if (pos1 > -1) {
    i = rank_index(rank, k, pos1);
    *betaj1 = beta[i];
  }
  if (pos2 > -1) {
    i = rank_index(rank, k, pos2);
    *betaj2 = beta[i];
  }

}

/* ## generate a beta type beta(a, b) value in the region of [x_min, x_max] */
/* as a increasing, the value gets close to x_max */
/* if a = b= 1, become the same as the uniform distribution */

double gen_beta(x_min, x_max, a, b)
double x_min, x_max, a, b;
{
  double t1, t2, x, y;
  t1 = MIN(x_min, x_max);
  t2 = MAX(x_min, x_max);
  /*x = gsl_ran_beta(rngptr, a, b);*/
  x = rbeta(a, b);
  y = t1 + (t2 -t1) * x;
  return(y);
}

/* rescaled the distribution */
double prob_beta(x, x_min, x_max, a, b)
     double x, x_min, x_max, a, b;
{
  double y, z, t1, t2;
  t1 = MIN(x_min, x_max);
  t2 = MAX(x_min, x_max);
  y = (x - t1) / (t2 - t1);
  /*z = gsl_ran_beta_pdf(y, a, b) / (t2 - t1);*/
  z = dbeta(y, a, b, 0)/(t2 - t1);
  return(z);
}


int fun_split(datastr)
struct data *datastr;
{
  int k=datastr->num_of_cluster, n=datastr->num_of_grp, counts_z[datastr->num_of_cluster], 
    counts_zprime[datastr->num_of_cluster+2], *Z=datastr->alloc_z;
  int i, ret=0, jstar, jminus, jplus, order[datastr->num_of_cluster], pos, Zprime[datastr->num_of_grp];
  int  *ptri, k0;
  double alphajminus, alphajplus, betajminus, betajplus, alphajstar, betajstar, 
    *alpha=datastr->alpha, *beta=datastr->beta;
  double **sim_mat=datastr->simi_mat;
  double betaj1, betaj2, logp_alloc, psi=datastr->psi_interaction, R, R1, R2, R3, R4, R5, num, den;
  double shape_split_a, shape_split_b, shape_merge_a, shape_merge_b;
  int **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor, km1;

  if (k == datastr->k_max) return(0);

  /* Get the cluster counts */
  cluster_counts(counts_z, k, Z, n);

  /* Make sure there is at least 1 cluster with more than 2* cluster_size, so each subcluster after the  */
  /* split could have at least cluster_size */
  /* Get k0 = the number of clusters with more than 2 * cluster_size subject in the old Z */
  k0 = 0;
  for (i=0, ptri=counts_z; i<k; i++, ptri++) {
    if (*ptri >= (2 * datastr->cluster_size)) {
      k0 = k0 + 1;
    }
  }
  if (k0 == 0) return(0);

  /* Get the ranks of beta */
  get_ranks(beta, k, order);

  /* Randomly pick a cluster with > 1 subject */
  km1 = k - 1;
  while (1) {
    /*jstar = gsl_rng_uniform_int(rngptr, k);*/
    jstar = uniform_int(0, km1);
    if (counts_z[jstar] >= (2 * datastr->cluster_size)) break;
  }

  jminus = k; /* Cluster numbers are from 0 to k-1 */
  jplus = k + 1; /* f can take k_max+1 values */ 
  alphajstar = alpha[jstar];
  betajstar = beta[jstar];
  pos = order[jstar];

  /* Get betaj1 and betaj2 */
  get_betaj12_split(beta, order, k, pos, &betaj1, &betaj2, datastr->beta_min, datastr->beta_max);

  /* Generate fj, betaj */
  /*  alphajminus = uniform(rngptr, datastr->alpha_min, alphajstar); */
  /* alphajplus = uniform(rngptr, alphajstar, datastr->alpha_max);*/
  /* betajminus = uniform(rngptr, betaj1, betajstar); */
  /* betajplus = uniform(rngptr, betajstar, betaj2); */

  shape_split_a = datastr->shape_split_a;
  shape_split_b = datastr->shape_split_b;
  shape_merge_a = datastr->shape_merge_a;
  shape_merge_b = datastr->shape_merge_b;
 
  alphajminus = gen_beta(datastr->alpha_min, alphajstar, shape_split_b, shape_split_a);
  alphajplus  = gen_beta(alphajstar, datastr->alpha_max, shape_split_a, shape_split_b);
  
  /* should be able to switch */
  if(bernoulli(0.5))
    {
      R1 = alphajminus;
      alphajminus = alphajplus;
      alphajplus = R1;
    }

  betajminus = gen_beta(betaj1, betajstar, shape_split_b, shape_split_a);
  betajplus = gen_beta(betajstar, betaj2,  shape_split_a, shape_split_b);

  /* Assign subjects to clusters and compute log(p_alloc) */
  ivec_ivec_init(Zprime, Z, n); /* Zprime must be initialized before calling log_p_alloc */
  logp_alloc = assign_split(datastr, alphajminus, alphajplus,  betajminus, betajplus, Zprime, jstar,
			    jminus, jplus);

  cluster_counts(counts_zprime, k+2, Zprime, n);
  /* Check that not all subjects were put in the same cluster */
  if (counts_zprime[jminus] < datastr->cluster_size || counts_zprime[jplus] < datastr->cluster_size)
    {  
      /* reject the split proposal */
	  return(0);
    }


  /* Calculate components of R (on log scale) */
  R1 = log_prob_1clus(datastr, alphajminus, betajminus, Zprime, jminus) + 
    log_prob_1clus(datastr, alphajplus, betajplus, Zprime, jplus) - 
    log_prob_1clus(datastr, alphajstar, betajstar, Z, jstar);

 

  R2 = log(pri_pd_alpha(alphajminus, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2)) +
       log(pri_pd_alpha(alphajplus, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2))  +
       log(pri_pd_beta(betajminus, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2)) +
       log(pri_pd_beta(betajplus, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2))- 
	   log(pri_pd_alpha(alphajstar, datastr->alpha_min, datastr->alpha_max, datastr->method_alpha, datastr->alpha_mu, datastr->alpha_sigma2)) -
	   log(pri_pd_beta(betajstar, datastr->beta_min, datastr->beta_max, datastr->method_beta, datastr->beta_mu, datastr->beta_sigma2));
 
  R3 = psi*(calc_Uz(sim_mat, Zprime, n, neighbor_id, num_neighbor) - calc_Uz(sim_mat, Z, n, neighbor_id, num_neighbor));

  num = prob_beta(alphajstar, alphajminus, alphajplus, shape_merge_a, shape_merge_b) *
    prob_beta(betajstar, betajminus, betajplus, shape_merge_a, shape_merge_b) /
    prob_beta(MIN(alphajminus, alphajplus),  datastr->alpha_min, alphajstar, shape_split_b, shape_split_a) /
    prob_beta(MAX(alphajminus, alphajplus), alphajstar, datastr->alpha_max, shape_split_a, shape_split_b) /
    prob_beta(betajminus, betaj1, betajstar, shape_split_b, shape_split_a) /
    prob_beta(betajplus, betajstar, betaj2,  shape_split_a, shape_split_b) * 2.0;  

  /* the reason of 2.0 at the end is for the uncertainty for which is alphajplus, which is alphajminus */
  den = k0*(datastr->d[k]) / (datastr->b[k-1] * k);
  R4 = log(num * den) - logp_alloc;

  /* note, d[k-1] is associated with clusters */
  /* den = (datastr->b[k-1])*(k)*fabs(alphajplus-alphajminus)*(betajplus-betajminus); */
  /* num = k0*(datastr->d[k])*(alphajstar-datastr->alpha_min)*(datastr->alpha_max-alphajstar)* */
  /*  (betajstar-betaj1)*(betaj2-betajstar);*/
 
  /*  R4 = log(num/den) - logp_alloc;*/
  


  R5 = R1 + R2 + R3 + R4;
  R = logR_to_R(R5);
  


  if (bernoulli(R)) {
    /* Accept the split */
    ivec_ivec_init(Z, Zprime, n);
    alpha[jminus] = alphajminus;
    alpha[jplus]  = alphajplus;
    beta[jminus] = betajminus;
    beta[jplus] = betajplus;
    /* Set num of clusters to k + 2, and let the check_z function make the correction */
    datastr->num_of_cluster = k + 2;
    check_z(datastr);
    ret = 1;
  }

  return(ret);
}







