#include "./header.h"

/* Function to get the next token */
static char * next_token(str, n, sep)
char *str, sep;
long n;
{
  int i;
  char *pc, c1;

  if (str == NULL) return(NULL);
  
  /* Loop till a sep character or end of string */
  for (i=0, pc=str; i<n; i++, pc++) {
    c1 = *pc;
    if (c1 == sep) {
      pc++;
      return(pc);
    } else if (c1 == END_OF_STR) {
      return(NULL);
    }
  }
  return(NULL);

} /* END: next_token */

/* Function to copy a token from a line read in to a field and adding '\0' */
static int copy_token(ret, str, sep, n)
char *ret; /* Return string */
char *str; /* Must start at the correct position */
char sep;
long n;
{
  int i, flag=2;
  char *pc, *pret, c1;
  
  /* The end-of-line character should appear before the end-of-string character */
 
  /*n = MIN(n, strlen(str));*/
  for (i=0, pc=str, pret=ret; i<n; i++, pc++, pret++) {
    c1 = *pc;

    if ((c1 != sep) && (c1 != END_OF_LINE) && (c1 != CAR_RETURN) && (c1 != EOF)) {
      *pret = *pc;
      flag = 1;
    } else {
       *pret = END_OF_STR;
       flag--;
       break;
    }
  }

  return(flag);

} /* END: copy_token */

/* Function to count a char in a string */
static int count_char_maxlen(str, c1, maxlen)
char *str, c1;
int *maxlen;
{
  int i, len, n, max, toklen;

  len = strlen(str);

  n = 0;
  max = -1;
  toklen = 0;
  for (i=0; i<len; i++) {
    toklen++;
    if (str[i] == c1) {
      n++;
      if (toklen > max) max = toklen;
      toklen = 0;
    }
  }
  if (!n) max = len;

  *maxlen = max;
  return(n);

} /* END: count_char */

/* Function to parse a delimited string into a double vector. Memory is allocated for the return vector */
static double * strIntoVec(str, sep, ret_n)
char *str, sep;
int *ret_n;
{
  char *token, cnum[101], *end;
  int i, n, maxlen;
  double *ret_vec;

  /* Get the number of sub-strings */
  n = count_char_maxlen(str, sep, &maxlen) + 1;

  maxlen = maxlen + 1;

  *ret_n = n;

  ret_vec = (double *) malloc(n*sizeof(double));
  CHECK_MEM(ret_vec);
  token = str;

  for (i=0; i<n; i++) {
    copy_token(cnum, token, sep, maxlen);
    ret_vec[i] = strtod(cnum, &end);
    token = next_token(token, maxlen, sep);
  } 

  return(ret_vec);

} /* END: strIntoVec */

static void get_option(parmfile, option, outstring)
char *parmfile, *option, *outstring;
{
  FILE *fpt;
  char line[MAXPARMLEN], *token, sep=',';
  int len, flag;
  size_t size;  

  size = sizeof(line);
  len = (int) strlen(option);

  /* Parameter file must be NAME,VALUE */
  fpt = open_r(parmfile);

  flag = 1;
  while (fgets(line, size, fpt) != NULL) {
    flag = strncmp(line, option, len);
    if (!flag) break;
  }
  fclose(fpt);
  if (flag) {
    printf("\n ERROR: The option name %s was not found in the parameter file. \n", option);
    exit(1);
  }

  /* Get the value */
  token = next_token(line, MAXPARMLEN, sep);
  if (token == NULL) {
    printf("\n ERROR: The value of option %s could not be retrieved from the parameter file. \n", option);
    exit(1);
  }
  flag = copy_token(outstring, token, sep, MAXPARMLEN);
  if (flag) {
    printf("\n ERROR: while copying value of option %s from the parameter file. \n", option);
    exit(1);
  }

} 

static void read_parms(datastr, parmfile, func)
struct data *datastr;
const char *parmfile;
int func; /* 0 = trans, 1 = trans_Uz */
{
  FILE *fpt;
  char line[MAXPARMLEN], **ptrc, *end, *pc;
  double x;  
  int np=38;
  int temp, pflag[np], i, *ptri, len[np], which, sum;
  char *parms[38] = {"alpha_max", "alpha_min", "beta_max", "beta_min", "cluster_size", "k_max",	\
                     "num_of_cluster",  "n_iter", "n_sep_out", "psi_max", "psi_min",\
		     "psi_k", "psi_theta", "random_seed", \
		     "sigma2_alpha", "sigma2_beta", "sigma2_psi", "sigma2_tao", "tao_max", "tao_min", \
                     "update_prop_w", "w_m", "method_psi", "split_option", "shape_split_a", "shape_split_b", \
                     "shape_merge_a", "shape_merge_b", "update_prop", "method_alpha", "method_beta", "method_tao",\
                     "alpha_mu", "alpha_sigma2", "beta_mu", "beta_sigma2", "tao_mu", "tao_sigma2"};
  size_t size;
  char sep=',';

  /* Initialize some variables */
  datastr->tao_min              = NULL;
  datastr->tao_max              = NULL;
  datastr->tao_mu               = NULL;
  datastr->tao_sigma2           = NULL;
  datastr->tao_min_n0           = 0;
  datastr->tao_max_n0           = 0;
  datastr->tao_mu_n0            = 0;
  datastr->tao_sigma2_n0        = 0;
  datastr->method_alpha         = -1;
  datastr->method_beta          = -1;
  datastr->method_tao           = -1;
  datastr->method_psi           = -1;
  datastr->psi_grid             = NULL;
  datastr->prob_psi_grid        = NULL;
  datastr->normalizing_psi_grid = NULL;


  size = sizeof(line);
  for (i=0, ptri=pflag; i<np; i++, ptri++) *ptri = 0;
  for (i=0, ptri=len, ptrc=parms; i<np; i++, ptri++, ptrc++) *ptri = (int) strlen(*ptrc); 

  fpt = open_r(parmfile);
  
  while (fgets(line, size, fpt) != NULL) {
    which = -1;
    for (i=0, ptrc=parms, ptri=len; i<np; i++, ptrc++, ptri++) {
      temp = strncmp(line, *ptrc, *ptri);
      if (!temp) {
        which = i; 
        break;
      }
    }

    if (which == -1) continue;
    pflag[which] = 1;
    pc = line + len[which] + 1;
    x = strtod(pc, &end);

    switch (which) {
      case 0:
        datastr->alpha_max = x;
        break;
      case 1:
        datastr->alpha_min = x;
        break;
      case 2:
        datastr->beta_max = x;
        break;
      case 3:
        datastr->beta_min = x;
        break;
      case 4:
        datastr->cluster_size = x;
        break;
      case 5:
        datastr->k_max = (int) x;
        break;
      case 6:
        datastr->num_of_cluster = (int) x;
        break;
      case 7:
        datastr->n_iter = (int) x;
        break;
      case 8:
        datastr->n_sep_out = (int) x;
        break;
      case 9:
        datastr->psi_max = x;
        break;
      case 10:
        datastr->psi_min = x;
        break;
      case 11:
        datastr->psi_k = x;
        break;
      case 12:
        datastr->psi_theta = x;
        break;
      case 13:
        datastr->random_seed = (int) x;
        break;
      case 14:
        datastr->sigma2_alpha = x;
        break;
      case 15:
        datastr->sigma2_beta = x;
        break;
      case 16:
        datastr->sigma2_psi = x;
        break;
      case 17:
        datastr->sigma2_tao = x;
        break;
      case 18:
        datastr->tao_max = strIntoVec(pc, sep, &datastr->tao_max_n0);
        break;
      case 19:
        datastr->tao_min = strIntoVec(pc, sep, &datastr->tao_min_n0);
        break;
      case 20:
        datastr->update_prop_w = x;
        break;
      case 21:
        datastr->w_m = (int) x;
        break;
      case 22:
        datastr->method_psi = (int) x;
        break;
      case 23:
        datastr->split_option = (int) x;
        break;
      case 24:
        datastr->shape_split_a = (int) x;
        break;
      case 25:
        datastr->shape_split_b = (int) x;
        break;
      case 26:
        datastr->shape_merge_a = (int) x;
        break;
      case 27:
        datastr->shape_merge_b = (int) x;
	break;
      case 28:
        datastr->update_prop = x;
        break;
      case 29:
        datastr->method_alpha = (int) x;
	break;
      case 30:
        datastr->method_beta = (int) x;
	break;
      case 31:
        datastr->method_tao = (int) x;
	break;
      case 32:
        datastr->alpha_mu = x;
	break;
      case 33:
        datastr->alpha_sigma2 = x;
	break;
      case 34:
        datastr->beta_mu = x;
	break;
      case 35:
        datastr->beta_sigma2 = x;
	break;
      case 36:
        datastr->tao_mu = strIntoVec(pc, sep, &datastr->tao_mu_n0);
	break;
      case 37:
        datastr->tao_sigma2 = strIntoVec(pc, sep, &datastr->tao_sigma2_n0);
	break;
      default:
        break;
    }
  }
  fclose(fpt);
  
  if (datastr->method_alpha == 0) {
    pflag[32] = 1;
    pflag[33] = 1;
  }
  if (datastr->method_alpha == 1) {
    pflag[0]  = 1;
    pflag[1]  = 1;
  }
  if (datastr->method_beta == 0) {
    pflag[34] = 1;
    pflag[35] = 1;
  }
  if (datastr->method_beta == 1) {
    pflag[2]  = 1;
    pflag[3]  = 1;
  }
  if (datastr->method_tao == 0) {
    pflag[36] = 1;
    pflag[37] = 1;
  }
  if (datastr->method_tao == 1) {
    pflag[18] = 1;
    pflag[19] = 1;
  }
  
  /* If func = 1 (trans_Uz function), most parms do not need to be specified. Set w_m = 1 */
  if (func == 1) {
    datastr->w_m = 1;
    for (i=0; i<5; i++) pflag[i] = 1;
    for (i=7; i<np; i++) pflag[i] = 1;
  }

  /* Check that all parms were specified */
  sum = 0;
  for (i=0, ptri=pflag; i<np; i++, ptri++) sum += *ptri; 
  if (sum < np) {
    printf("\n ERROR: The following (case sensitive) parameters were not specified: \n");
    for (i=0, ptri=pflag, ptrc=parms; i<np; i++, ptri++, ptrc++) {
      if (!*ptri) printf("%s \n", *ptrc);
    }
    exit(1);
  }
} 

/* Function to read file for method_psi = 2 */ 
void readPsiFile(datastr, filename)
struct data *datastr;
char *filename;
{
  int i, nr, ret;
  size_t size;
  char line[MAXPARMLEN], *end, *token, cnum[101], sep=' ';
  FILE *fpt;
  double *grid, *prob, *norm;

  /* Get the number of rows in the file */
  nr = getNrows(filename);
  if (!nr) {
    printf("ERROR: No rows in psi_file\n");
    exit(1);
  }
  datastr->n_psi_grid = nr;

  datastr->psi_grid = (double *) malloc(nr*sizeof(double));
  CHECK_MEM(datastr->psi_grid);
  datastr->prob_psi_grid = (double *) malloc(nr*sizeof(double));
  CHECK_MEM(datastr->prob_psi_grid);
  datastr->normalizing_psi_grid = (double *) malloc(nr*sizeof(double));
  CHECK_MEM(datastr->normalizing_psi_grid);

  grid = datastr->psi_grid;
  prob = datastr->prob_psi_grid;
  norm = datastr->normalizing_psi_grid;
  size = sizeof(line);

  fpt = open_r(filename);
  i = 0;
  while (fgets(line, size, fpt) != NULL) {

    token = line;

    /* Column 1: psi_grid */
    ret = copy_token(cnum, token, sep, 100);
    grid[i] = strtod(cnum, &end);

    /* Column 2: prob_psi_grid */
    token = next_token(token, 100, sep);
    ret = copy_token(cnum, token, sep, 100);
    prob[i] = strtod(cnum, &end);

    /* Column 3: normalizing_psi_grid */
    token = next_token(token, 100, sep);
    ret = copy_token(cnum, token, sep, 100);
    norm[i] = strtod(cnum, &end);

    i++;
  }

  fclose(fpt);

} /* END: readPsiFile */

/****************************************************************************************/
/**************** Main functions ********************************************************/
/****************************************************************************************/
void trans(pargc, argv, ret)
 int *pargc;
 char **argv;
 int *ret;
{
  char *filename, *sim_mat_file;
  struct data datastr;
  int i, jj, k, *ptri, km1, nsubm1, n_sep_out, sep_index, ncovar;
  double *ptrd;
  FILE *fpt1, *fpt2, *fpt3, *fpt4, *fpt5, *fpt6, *fpt7=NULL;
  int pathlen=MAXPATHLEN, psi_method;
  

  *ret = 1;
  filename = (char *) malloc(pathlen*sizeof(char));
  CHECK_MEM(filename);
  sim_mat_file = (char *) malloc(pathlen*sizeof(char));
  CHECK_MEM(sim_mat_file);

  /* Read the file of parameters and put them in the data structure */
  read_parms(&datastr, argv[0], 0);
 
  /* For method_psi = 2 */
  if (datastr.method_psi == 2) {
    get_option(argv[0], "psi_file", filename);
    readPsiFile(&datastr, filename);
  }

  /* Get the path to the input data set and similarity matrix */
  get_option(argv[0], "data", filename);
  get_option(argv[0], "sim_matrix", sim_mat_file);

  /* Initialize the data structure. This functions allocates memory, reads in the data, and initializes
     other variables. */
  init(&datastr, filename, sim_mat_file);

  free(sim_mat_file);

  /****************************************************************************/
  /****************************************************************************/
  /****************************************************************************/

  nsubm1 = datastr.num_of_grp - 1;
  n_sep_out = datastr.n_sep_out;
  sep_index = 0;
  ncovar = datastr.num_of_covars;
  psi_method = datastr.method_psi;

  /* Open output data sets, k, psi, f, beta, z */
  /*
  get_option(argv[0], "out_k", filename);
  fpt1 = fopen(filename, "w");
  */
  get_option(argv[0], "out_psi", filename);
  fpt2 = fopen(filename, "w");
  get_option(argv[0], "out_alpha", filename);
  fpt3 = fopen(filename, "w");
  get_option(argv[0], "out_beta", filename);
  fpt4 = fopen(filename, "w");
  get_option(argv[0], "out_alloc_z", filename);
  fpt5 = fopen(filename, "w");
  get_option(argv[0], "out_dev_i", filename);
  fpt6 = fopen(filename, "w");
  get_option(argv[0], "out_tao", filename);
  if (ncovar > 0) {
    fpt7 = fopen(filename, "w"); 
    free(filename);
  }
   
  for (i=1; i<=datastr.n_iter; i++) {

    /* Update the interaction parameter */
    if (datastr.num_of_cluster > 1)
    {
      if (psi_method == 2) {
        jj = update_psi_discrete(&datastr);
      } else if (datastr.psi_max > (datastr.psi_min + 0.000001)) {
	 /* otherwise, do not update psi */
        jj = update_psi(&datastr);      
      }
      jj = update_z(&datastr);
    }
     
    if (datastr.num_of_covars > 0) { 
      jj = update_alpha_beta_tao(&datastr);
    } else {
      jj = update_alpha_beta(&datastr);
    }
 
    sep_index++;
    if (sep_index == n_sep_out) {
      sep_index = 0;

      /* Write out results */
      k = datastr.num_of_cluster;
      km1 = k - 1;
      /* k */
      /*
      fprintf(fpt1, "%d\n", k);
      jj = fflush(fpt1);
      */

      /* psi */
      fprintf(fpt2, "%g\n", datastr.psi_interaction);
      jj = fflush(fpt2);

      /* alpha */
      for (jj=0, ptrd=datastr.alpha; jj<km1; jj++, ptrd++) fprintf(fpt3, "%g ", *ptrd);
      fprintf(fpt3, "%g\n", *ptrd);
      jj = fflush(fpt3);

      /* beta */ 
      for (jj=0, ptrd=datastr.beta; jj<km1; jj++, ptrd++) fprintf(fpt4, "%g ", *ptrd);
      fprintf(fpt4, "%g\n", *ptrd);
      jj = fflush(fpt4);

      /* z */
      for (jj=0, ptri=datastr.alloc_z; jj<nsubm1; jj++, ptri++) fprintf(fpt5, "%d ", *ptri);
      fprintf(fpt5, "%d\n", *ptri);
      jj = fflush(fpt5);

      /* dev_i */
      fprintf(fpt6, "%g\n", datastr.dev_i);
      jj = fflush(fpt6);

      /* tao */ 
      if (ncovar > 0) {
        for (jj=0, ptrd=datastr.tao; jj<ncovar; jj++, ptrd++) fprintf(fpt7, "%g ", *ptrd);
	 fprintf(fpt7, "\n");
        jj = fflush(fpt7);
      }
    }
  } 
  /*fclose(fpt1);*/
  fclose(fpt2);
  fclose(fpt3);
  fclose(fpt4);
  fclose(fpt5);
  fclose(fpt6);
  if (ncovar > 0) fclose(fpt7);

  free_memory(&datastr);
  *ret = 0;
  return;
}

/* Function to read in data and set up the data structure */
static void setUp0(config, datastr, which)
char *config;
struct data *datastr;
int which;  /* 1 = trans_Uz */
{
  char *filename, *sim_mat_file;
  int pathlen=MAXPATHLEN;
  
  filename = (char *) malloc(pathlen*sizeof(char));
  CHECK_MEM(filename);
  sim_mat_file = (char *) malloc(pathlen*sizeof(char));
  CHECK_MEM(sim_mat_file);

  /* Read the file of parameters and put them in the data structure */
  read_parms(datastr, config, 1);
 
  if (which == 1) {
    datastr->method_psi = 0;
  }

  /* Get the path to the input data set and similarity matrix */
  get_option(config, "data", filename);
  get_option(config, "sim_matrix", sim_mat_file);

  /* Initialize the data structure. This functions allocates memory, reads in the data, and initializes
     other variables. */
  init(datastr, filename, sim_mat_file);
  
  free(sim_mat_file);
  free(filename);

} /* END: setUp0 */

void trans_Uz(pargc, argv, p_psi_fix, p_n_iter, p_n_sep_out, Uz)
int *pargc;
char **argv;
double *p_psi_fix;
int *p_n_iter, *p_n_sep_out;
double *Uz;
{
  struct data datastr;
  int i, n_sep_out, sep_index, ngrp, Uz_index=0;
  double temp, psi_fix;
  int *z_in, *z_out, n_iter;
  
  setUp0(argv[0], &datastr, 1);

  /****************************************************************************/
  /****************************************************************************/
  /****************************************************************************/

  sep_index = 0;
  n_iter = *p_n_iter;
  n_sep_out = *p_n_sep_out;

  ngrp    = datastr.num_of_grp;  
  z_out = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(z_out);
  z_in = (int *) malloc(ngrp*sizeof(int));
  CHECK_MEM(z_in);
                                                                                                                                                                                           
  ini_gen_alloc(z_in, ngrp, datastr.num_of_cluster);

  /* Uz is a double vector with length of datastr.n_iter/n_sep_out */
  psi_fix = *p_psi_fix;

  for (i=1; i<=n_iter; i++)
  {
    gen_alloc_cond1(z_out,  psi_fix, datastr.num_of_cluster, datastr.simi_mat, z_in, ngrp, datastr.neighbor_id,
                    datastr.num_neighbor);      
    ivec_ivec_init(z_in, z_out, ngrp);
   
    sep_index++;

    if (sep_index == n_sep_out) {
      sep_index = 0;

      /* do the following for every n_sep_out iteration */
      temp = calc_Uz(datastr.simi_mat, z_in, ngrp, datastr.neighbor_id, datastr.num_neighbor);    
      Uz[Uz_index++] = temp;
    } 
  }

  free(z_out);
  free(z_in);
  free_memory(&datastr);
  
  return;
}

static void normC_SAMC_main(datastr, psi_vec, nclus, num_grid, num_iter, gain_factor_t0, theta_start, hist_mat)
struct data *datastr;
double *psi_vec;
int  nclus, num_grid, num_iter;
double gain_factor_t0, *theta_start;
double **hist_mat;
{
  int i, j, nm1, nm2, i_grid_0, i_grid_1, i_grid_new, ngrp=datastr->num_of_grp, iter;
  double *pd, temp, tie_vec[num_grid], **Q, **sim_mat=datastr->simi_mat, logr, r, vec[3];  
  int **neighbor_id=datastr->neighbor_id, *num_neighbor=datastr->num_neighbor, z0[ngrp], z1[ngrp], *pi1, *pi2;

  Q = dblMatrix_alloc(num_grid, num_grid);

  /* Initialize */
  for (i=0; i<num_grid; i++) {
    pd    = hist_mat[i];
    pd[0] = psi_vec[i];
    pd[1] = theta_start[i];
    pd[2] = 0.0;
  }
  temp = 1.0/num_grid;
  for (i=0, pd=tie_vec; i<num_grid; i++, pd++) *pd = temp;

  /* Q */
  for (i=0; i<num_grid; i++) {
    for (j=0, pd=Q[i]; j<num_grid; j++, pd++) *pd = 0.0;
  }
  nm1     = num_grid - 1;
  nm2     = num_grid - 2;
  temp    = 1.0/3.0;
  Q[0][0] = Q[nm1][nm1] = 2.0/3.0;
  Q[0][1] = Q[nm1][nm2] = temp;
  for (i=1; i<nm1; i++) Q[i][i-1] = Q[i][i] = Q[i][i+1] = temp;

  /* Uniform integer btween 0 and num_grid-1 */
  i_grid_0 = uniform_int(0, nm1);

  /* Generate z0 */
  ini_gen_alloc(z0, ngrp, nclus);

  for (iter=1; iter<=num_iter; iter++) {
    if (i_grid_0 == 0) {
      vec[0] = 0.0;
      vec[1] = Q[0][0];
      vec[2] = Q[0][1];
    } else if (i_grid_0 == nm1) {
      vec[0] = Q[nm1][nm2];
      vec[1] = Q[nm1][nm1];
      vec[2] = 0.0;
    } else {
      vec[0] = Q[i_grid_0][i_grid_0 - 1];
      vec[1] = Q[i_grid_0][i_grid_0];
      vec[2] = Q[i_grid_0][i_grid_0 + 1];
    }
    i_grid_1 = sample1(vec, 3) + i_grid_0 - 1;

    if (i_grid_1 == i_grid_0) {
      gen_alloc_cond1(z1, psi_vec[i_grid_0], nclus, sim_mat, z0, ngrp, neighbor_id, num_neighbor);
      i_grid_new = i_grid_1;
      for (i=0, pi1=z0, pi2=z1; i<ngrp; i++, pi1++, pi2++) *pi1 = *pi2;
    } else {
      temp = calc_Uz(sim_mat, z1, ngrp, neighbor_id, num_neighbor);    
      logr = hist_mat[i_grid_0][1] - hist_mat[i_grid_1][1] + (psi_vec[i_grid_1] - psi_vec[i_grid_0])*temp +
      log(Q[i_grid_1][i_grid_0]) - log(Q[i_grid_0][i_grid_1]);

      r = logR_to_R(logr);
      if (bernoulli(r)) {
        i_grid_new = i_grid_1;
      } else {
        i_grid_new = i_grid_0;
      }
    }

    r = gain_factor_t0/MAX(gain_factor_t0, iter);
    for (i=0; i<num_grid; i++) {
      if (i != i_grid_new) {
        hist_mat[i][1] -= r*tie_vec[i];
      } else {
        hist_mat[i][1] += r*(1.0 - tie_vec[i]); 
      }
    }

    hist_mat[i_grid_new][2] += 1.0;
    i_grid_0 = i_grid_new;

  } /* END: for (iter=0; iter<num_iter; iter++) */

  dblMatrix_free(Q, num_grid);

} /* END: normC_SAMC_main */

void normC_SAMC(pargc, argv, psi_vec, nclus, num_grid, num_iter, gain_factor_t0, theta_start, hist_mat)
int *pargc;
char **argv;
double *psi_vec;
int *nclus, *num_grid, *num_iter;
double *gain_factor_t0, *theta_start;
double *hist_mat;
{
  struct data datastr;
  int i, j, ngrid=*num_grid;
  double **mat, *pd;
  
  setUp0(argv[0], &datastr, 1);

  mat = dblMatrix_alloc(ngrid, 3);
  normC_SAMC_main(&datastr, psi_vec, *nclus, ngrid, *num_iter, *gain_factor_t0, theta_start, mat);

  /* hist_mat is a vector: first 3 elements is row 1, next 3 elements is row 2, etc */
  pd = hist_mat;
  for (i=0; i<ngrid; i++) {
    for (j=0; j<3; j++) *pd++ = mat[i][j];
  } 

  dblMatrix_free(mat, ngrid);
  return;

} /* END: normC_SAMC */ 

